# -*- coding: utf-8 -*-

# Copyright (C) 2019 Christoph G. Hoffmann. All rights reserved.

# This file is part of mjoindices

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Contact: christoph.hoffmann@uni-greifswald.de

"""
This module provides a simple 1-dim filtering algorithm, which can be used during the PC calculation instead of the
full 2-dim Wheeler-Kiladis-Filter.

This module is not intended to be used stand-alone outside the OMI context, as it has only been extensively 
tested for the specific OMI filtering conditions.

Hence, there is usually no need for the users of the mjoindices package to call functions of this module themselves.
Instead, they probably want to use the module :py:mod:`mjoindices.omi.omi_calculator` directly.
"""

import numpy as np
import scipy
import scipy.fftpack

import mjoindices.olr_handling as olr


def filter_olr_for_mjo_pc_calculation_1d_spectral_smoothing(olrdata: olr.OLRData) -> olr.OLRData:
    """
    Filters OLR data temporally using a 1d Fourier transform filter.

    The temporal filtering constants are chosen to meet the values in the description by :ref:`refKiladis2014`.

    :param olrdata: The original OLR data

    :return: The filtered OLR.
    """
    return filter_olr_temporally_1d_spectral_smoothing(olrdata, 20., 96.)


def filter_olr_temporally_1d_spectral_smoothing(olrdata: olr.OLRData, period_min: float, period_max: float) -> olr.OLRData:
    """
    Filters OLR data temporally using a 1d Fourier transform filter.

    :param olrdata: The original OLR data
    :param period_min: Temporal filter constant: Only greater periods (in days) remain in the data.
    :param period_max: Temporal filter constant: Only lower periods (in days) remain in the data.

    :return: The filtered OLR.
    """
    filteredOLR = np.empty(olrdata.olr.shape)
    time_spacing = (olrdata.time[1] - olrdata.time[0]).astype('timedelta64[s]') / np.timedelta64(1, 'D')  # time spacing in days
    for idx_lat in range(0, olrdata.olr.shape[1]):
        for idx_lon in range(0, olrdata.olr.shape[2]):
            tempolr = np.squeeze(olrdata.olr[:, idx_lat, idx_lon])
            filteredOLR[:, idx_lat, idx_lon] = _perform_spectral_smoothing(tempolr, time_spacing, period_min, period_max)
    return olr.OLRData(filteredOLR, olrdata.time, olrdata.lat, olrdata.long)


def _perform_spectral_smoothing(y, dt, lower_cutoff, higher_cutoff):
    """
    Applies a 1d Fourier Transform filter to the vector y.

    :param y: The data to filter
    :param dt: The spacing of the data
    :param lower_cutoff: Filter constant: Only greater periods (same units as dt) remain in the data.
    :param higher_cutoff: Filter constant: Only lower periods (same units as dt) remain in the data.

    :return: The filtered vector
    """
    N = y.size
    w = scipy.fftpack.rfft(y)
    f = scipy.fftpack.rfftfreq(N, dt)
    P = 1 / f
    w2 = w.copy()
    w2[P < lower_cutoff] = 0
    w2[P > higher_cutoff] = 0
    y2 = scipy.fftpack.irfft(w2)
    return y2



# -*- coding: utf-8 -*-

# Copyright (C) 2019 Christoph G. Hoffmann. All rights reserved.

# This file is part of mjoindices

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Contact: christoph.hoffmann@uni-greifswald.de

"""
This module provides basic functionality to handle PC data, which is a basic output of the OMI calculation.
"""

import datetime
from pathlib import Path

import numpy as np
import pandas as pd


class PCData:
    """
    Class as a container for the principal component (PC) data.

    The basic PC computation :func:`mjoindices.omi.omi_calculator.calculate_pcs_from_olr` will return an object of this
    class as a major result of this package.

    :param time: Array containing the :class:`numpy.datetime64` dates.
	:param period: Array containing the :class:`pandas.Period` dates.
    :param pc1: Array containing the values of PC1 (has to be of same length as the time array).
    :param pc2: Array containing the values of PC2 (has to be of same length as the time array).
    """

    # ToDo: Check if we need some deepcopy or writable=false here
    def __init__(self, time: np.ndarray, pc1: np.ndarray, pc2: np.ndarray) -> None:
        """
        Initialization with all necessary variables.
        """
        if pc1.size != time.size:
            raise ValueError('Length of the first PC time series does not fit to the length of the '
                             'time grid')
        if pc2.size != time.size:
            raise ValueError('Length of the second PC time series does not fit to the length of the '
                             'time grid')
        self._time = time.copy()
        self._pc1 = pc1.copy()
        self._pc2 = pc2.copy()

    @property
    def time(self) -> np.ndarray:
        """
        The time grid of the PC time series as array of :class:`numpy.datetime64` elements.
        """
        return self._time

    @property
    def pc1(self) -> np.ndarray:
        """
        The time series of the PC1 values.
        """
        return self._pc1

    @property
    def pc2(self) -> np.ndarray:
        """
        The time series of the PC2 values.
        """
        return self._pc2

    def save_pcs_to_txt_file(self, filename: Path) -> None:
        """
        Saves the computed PCs to a text file.

        Please note that the file format is not exactly that of the original data files. However, a
        suitable reader is available in this module for both formats
        (:func:`mjoindices.principal_components.load_pcs_from_txt_file` and
        :func:`mjoindices.principal_components.load_original_pcs_from_txt_file`).

        :param filename: The full filename.
        """
        df = pd.DataFrame({"Date": self._time, "PC1": self._pc1, "PC2": self._pc2})
        df.to_csv(filename, index=False, float_format="%.5f")


def load_pcs_from_txt_file(filename: Path) -> PCData:
    """
    Loads the PCs of OMI, which were previously saved with this package
    (:func:`mjoindices.principal_components.PCData.save_pcs_to_txt_file`).

    :param filename: Path to the PC file.

    :return: The PC data.
    """
    df = pd.read_csv(filename, sep=',', parse_dates=[0], header=0)
    dates = df.Date.values
    pc1 = df.PC1.values
    pc2 = df.PC2.values
    return PCData(dates, pc1, pc2)


def load_original_pcs_from_txt_file(filename: Path) -> PCData:
    """
    Loads the PCs of OMI, which are stored in the original file format.
    For example, the following file can be loaded: https://www.esrl.noaa.gov/psd/mjo/mjoindex/omi.1x.txt

    Note that the present software package stores the PCs slightly different. Those files can be loaded with
    :func:`load_pcs_from_txt_file`.

    :param filename: Path to the PC file.

    :return:  The original PC data.
    """
    my_data = np.genfromtxt(filename)
    dates_temp = []
    for i in range(0, my_data.shape[0]):
        dates_temp.append(
            datetime.datetime(my_data[i, 0].astype(int), my_data[i, 1].astype(int), my_data[i, 2].astype(int)))
    dates = np.array(dates_temp, dtype='datetime64')
    pc1 = np.array(my_data[:, 4])
    pc2 = np.array(my_data[:, 5])
    return PCData(dates, pc1, pc2)




"""
Contains the post-processing of the EOFs as described in :ref:`refWeidman2022`. This is an alternative post-processing
approach and does not lead to the same results as shown in :ref:`refKiladis2014`. It reduces noise and avoids
potential degeneracy issues.

The post-processing procedure follows the below steps:

#. Corrects spontaneous sign changes in the EOFs (same as the original procedure)
#. Projects EOFs at DOY = n-1 onto EOF space for DOY = n. This is done to reduce spurious oscillations between EOFs on sequential days
#. Rotate the projected EOFs by 1/366 (or 1/365) per day to ensure continuity across January to December
#. Renormalize the EOFs to have a length of 1 (this is a small adjustment to account for small numerical errors).

.. seealso:: :py:mod:`mjoindices.omi.postprocessing_original_kiladis2014`

"""

from typing import Tuple
import inspect

import numpy as np
import warnings
import importlib

import mjoindices.empirical_orthogonal_functions as eof
import mjoindices.omi.postprocessing_original_kiladis2014 as pp_kil2014
import mjoindices.tools as tools


def post_process_eofs_rotation(eofdata: eof.EOFDataForAllDOYs, sign_doy1reference: bool = True) -> eof.EOFDataForAllDOYs:
    """
    Executes the complete post-processing of a series of EOF pairs for all DOYs according to the alterntive procedure
    described by :ref:`refWeidman2022`.

    Postprocessing includes an alignment of EOF signs and a rotation algorithm that rotates the EOFs
    in three steps:

    #. Projects EOFs at DOY = n-1 onto EOF space for DOY = n. This is done to reduce spurious oscillations between EOFs
       on sequential days
    #. Rotate the projected EOFs by 1/366 (or 1/365) of the December --> January discontinuity each day to ensure 
       continuity across December to January
    #. Renormalize the EOFs to have a length of 1 (this is a small adjustment to account for small numerical errors).

    See documentation of the methods :py:func:`~mjoindices.omi.postprocessing_original_kiladis2014.correct_spontaneous_sign_changes_in_eof_series`
    for EOF sign flipping

    Note that it is recommended to use the function :py:func:`~mjoindices.omi.omi_calculator.calc_eofs_from_olr` to cover
    the complete algorithm.

    :param eofdata: The EOF series, which should be post processed.
    :param sign_doy1reference: See :py:func:`~mjoindices.omi.postprocessing_original_kiladis2014.correct_spontaneous_sign_changes_in_eof_series`.

    :return: The postprocessed series of EOFs

    """
    
    pp_eofs = pp_kil2014.correct_spontaneous_sign_changes_in_eof_series(eofdata, doy1reference=sign_doy1reference)
    rot_eofs = rotate_eofs(pp_eofs)
    norm_eofs = normalize_eofs(rot_eofs)
    
    return norm_eofs


def rotate_eofs(orig_eofs: eof.EOFDataForAllDOYs) -> eof.EOFDataForAllDOYs:
    """
    Rotate EOFs at each DOY to

    #. align with the EOFs of the previous day and
    #. be continuous across December to January boundary.

    Described more in detail in :py:func:'post_process_eofs_rotation'

    :param orig_eofs: Calculated EOFs, which already has algined signs between neighboring DOYs.

    :return: set of rotated EOFs.
    """

    delta = calculate_angle_from_discontinuity(orig_eofs)

    return rotate_each_eof_by_delta(orig_eofs, delta)


def rotation_matrix(delta):
    """
    Creates a 2d rotation matrix for corresponding delta.

    :param delta: Scalar angle, in radians, of desired rotation.

    :return: 2x2 rotation matrix that can be used to rotate an Nx2 matrix in the x-y plane counterclockwise 
        by delta.

    """
    return np.array([[np.cos(delta), -np.sin(delta)],[np.sin(delta), np.cos(delta)]])


def calculate_angle_from_discontinuity(orig_eofs: eof.EOFDataForAllDOYs):
    """
    Project the matrix to align with the EOFs from the previous DOY and calculate the resulting
    discontinuity between January 1 and January 1 after one year of projections. Divide by the number of days 
    in one year to determine the discontinuity used for rotation. 

    :param orig_eofs: calculated EOFs, after signs have been changed via spontaneous_sign_changes

    :return: float of angular discontinuity between EOF1 on DOY1 and EOF1 on DOY1 after a full year of 
        projection, divided by the length of the year.

    """

    list_of_doys = tools.doy_list(orig_eofs.no_leap_years)
    doy1 = orig_eofs.eofdata_for_doy(1)

    ndoys = len(list_of_doys)
    
    # set DOY1 initialization
    rots = np.array([doy1.eof1vector, doy1.eof2vector])

    # project onto previous day
    for d in list_of_doys:
        if d+1 > ndoys: # for last day in cycle, return to January 1
            doyn = orig_eofs.eofdata_for_doy(1)
        else:
            doyn = orig_eofs.eofdata_for_doy(d+1)

        B = np.array([doyn.eof1vector, doyn.eof2vector]).T 
        A = np.array([rots[0,:], rots[1,:]]).T
    
        rots = np.matmul(np.matmul(B, B.T),A).T
    
    # calculate discontinuity between Jan 1 and Jan 1 at end of rotation cycle
    discont = angle_btwn_vectors(doy1.eof1vector, rots[0,:])

    # determine whether to rotate clockwise or counterclockwise, based on angle of E1 from projected
    # E2 and angle of E2 from projected E1
    cross_angle = np.dot(doy1.eof1vector, rots[1,:])/(np.linalg.norm(doy1.eof1vector)*np.linalg.norm(rots[1,:]))
    if cross_angle <= 0:
        return -discont/ndoys
    else: 
        return discont/ndoys


def rotate_each_eof_by_delta(orig_eofs: eof.EOFDataForAllDOYs, delta: float) -> eof.EOFDataForAllDOYs:
    """
    Use delta calculated by optimization function to rotate original EOFs by delta.
    First projects EOFs from DOY n-1 onto EOF space for DOY n, then rotates projected
    EOFs by small angle delta. 

    :param orig_eofs: calculated EOFs, signs have been changed via spontaneous_sign_changes
    :param delta: scalar by which to rotate EOFs. Calculated as the angular discontinuity between EOF1 on DOY1 
        and EOF1 on DOY1 after a full year of projection, divided by the length of the year.

    :return: new EOFdata with projected and rotated EOFs.

    """

    R = rotation_matrix(delta)

    doy1 = orig_eofs.eofdata_for_doy(1)   
    list_of_doys = tools.doy_list(orig_eofs.no_leap_years)
    
    eofdata_rotated = []
    eofdata_rotated.append(doy1) # first doy is unchanged

    # set DOY1 initialization
    rots = np.array([doy1.eof1vector, doy1.eof2vector])

    # project onto previous day and rotate 
    for d in list_of_doys[1:]:
        doyn = orig_eofs.eofdata_for_doy(d)

        B = np.array([doyn.eof1vector, doyn.eof2vector]).T 
        A = np.array([rots[0,:], rots[1,:]]).T
    
        rots = np.matmul(np.matmul(np.matmul(B, B.T),A),R).T

        # create new EOFData variable for rotated EOFs
        eofdata_rotated.append(eof.EOFData(doyn.lat, doyn.long, 
                                np.squeeze(rots[0,:]), 
                                np.squeeze(rots[1,:]),
                                explained_variances=doyn.explained_variances,
                                eigenvalues=doyn.eigenvalues,
                                no_observations=doyn.no_observations))

    return eof.EOFDataForAllDOYs(eofdata_rotated, orig_eofs.no_leap_years)


def normalize_eofs(orig_eofs: eof.EOFDataForAllDOYs) -> eof.EOFDataForAllDOYs:
    """
    Normalize all EOFs to have a magnitude of 1.

    :param eofdata: The rotated EOF series.

    :return: normalized EOFdata for all days.
    """

    list_of_doys = tools.doy_list(orig_eofs.no_leap_years)

    eofdata_normalized = []

    for d in list_of_doys:

        doyn = orig_eofs.eofdata_for_doy(d)
        eof1_norm = doyn.eof1vector/np.linalg.norm(doyn.eof1vector)
        eof2_norm = doyn.eof2vector/np.linalg.norm(doyn.eof2vector) 

       # create new EOFData variable for rotated EOFs
        eofdata_normalized.append(eof.EOFData(doyn.lat, doyn.long, 
                                            np.squeeze(eof1_norm), 
                                            np.squeeze(eof2_norm),
                                            explained_variances=doyn.explained_variances,
                                            eigenvalues=doyn.eigenvalues,
                                            no_observations=doyn.no_observations)) 

    return eof.EOFDataForAllDOYs(eofdata_normalized, orig_eofs.no_leap_years) 


def angle_between_eofs(reference: eof.EOFData, target=eof.EOFData):
    """
    Calculates angle between two EOF vectors to determine their "closeness" :math:`theta = arccos(t . r / (||r||*||t||))`.

    :param reference: The reference-EOFs. This is usually the EOF pair of the previous or "first" DOY.
    :param target: The EOF data from the target DOY.

    :return: A tuple of the angles between the reference and target EOFs for both EOF1 and EOF2
    """

    angle1 = angle_btwn_vectors(reference.eof1vector, target.eof1vector)
    angle2 = angle_btwn_vectors(reference.eof2vector, target.eof2vector)

    return (angle1, angle2)


def angle_btwn_vectors(vector1, vector2):
    """
    Calculates the angle between vectors, :math:`theta = arccos(t . r / (||r||*||t||))`.

    :param vector1: 1d vector, generally corresponding to EOF1 or 2 from some DOY
    :param vector2: 1d vector, generally corresponding to EOF1 or 2 from a different DOY

    :return: scalar angle between vectors 1 and 2, in radians. 
    """

    return np.arccos(np.clip(np.dot(vector1, vector2)
                             /(np.linalg.norm(vector1)*np.linalg.norm(vector2)),-1.,1.))



# -*- coding: utf-8 -*-

# Copyright (C) 2019 Christoph G. Hoffmann. All rights reserved.

# This file is part of mjoindices

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Contact: christoph.hoffmann@uni-greifswald.de

"""
This module provides several methods that help evaluate the agreement of the OMI calculation by this package
and the original calculation by :ref:`refKiladis2014`.
It is probably not of major relevance for the user of this package.
"""

import typing
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import matplotlib.cm
import re
import warnings

import mjoindices.empirical_orthogonal_functions as eof
import mjoindices.principal_components as pc
import mjoindices.tools as tools
from pandas.plotting import register_matplotlib_converters

register_matplotlib_converters()


def compute_vector_difference_quantity(ref_vec: np.ndarray, vec: np.ndarray, percentage: bool = True) -> np.ndarray:
    """
    Calculates a standardized difference between two vectors.

    :param ref_vec: The reference vector.
    :param vec: The vector to validate.
    :param percentage: If ``True``: calculation results will be in percent of the mean of the absolute reference values.

    :return: A vector containing the difference.
    """
    result = vec - ref_vec
    if percentage is True:
        result = result / np.mean(np.abs(ref_vec)) * 100
    return result


def calc_vector_agreement(ref_vec: np.ndarray, vec: np.ndarray, percentage: bool = True, do_print: bool = False) -> typing.Tuple:
    """
    Calculates extended comparison statistics between two vectors.

    :param ref_vec: The reference vector.
    :param vec: The vector to validate.
    :param percentage: If ``True``: calculation results will be in percent of the mean of the absolute reference values.
    :param do_print: If ``True``, some statistical values will we printed to the console.

    :return: A tuple containing values for the following quantities: correlation, mean of the differences, standard
        deviation of the differences, and percentiles of the absolute differences for 68%, 95%, and 99%.
    """
    if not np.all(ref_vec.size == vec.size):
        raise ValueError("Vectors do not have the same lengths.")
    corr = np.corrcoef(ref_vec, vec)[0, 1]

    diff_vec = compute_vector_difference_quantity(ref_vec, vec, percentage=percentage)
    diff_mean = np.mean(diff_vec)
    diff_std = np.std(diff_vec)
    diff_abs_percent68 = np.percentile(np.abs(diff_vec), 68.)
    diff_abs_percent95 = np.percentile(np.abs(diff_vec), 95.)
    diff_abs_percent99 = np.percentile(np.abs(diff_vec), 99)

    if do_print:
        print("CorrelationCoefficient: %1.4f" % corr)
        print("Mean of difference: %1.4f" % diff_mean)
        print("Stddev. of difference: %1.4f" % diff_std)
        print("68%% Percentile (abs. value of differences: %1.4f" % diff_abs_percent68)
        print("95%% Percentile (abs. value of differences: %1.4f" % diff_abs_percent95)
        print("99%% Percentile (abs. value of differences: %1.4f" % diff_abs_percent99)

    return corr, diff_mean, diff_std, diff_vec, diff_abs_percent68, diff_abs_percent95, diff_abs_percent99


def calc_comparison_stats_for_eofs_all_doys(eofs_ref: eof.EOFDataForAllDOYs,
                                            eofs: eof.EOFDataForAllDOYs,
                                            eof_number,
                                            exclude_doy366: bool = False,
                                            percentage: bool=False,
                                            do_print: bool = False) -> typing.Tuple:
    """
    Calculates extended comparison statistics between calculated EOFs and reference EOFs.

    :param recalc_eof: The EOFs to validate.
    :param orig_eof: The reference EOFs.
    :param exclude_doy366: If ``True``, DOY 366 will be included in the calculation (sometimes worse agreement depending on
        the leap year treatment mode).
    :param percentage: If ``True``: calculation results will be in percent of the mean of the absolute reference values.
    :param do_print: If ``True``, some characteristic values will we printed to the console.

    :return: A tuple containing arrays with 366 or 365 elements each for the following quantities: correlation, mean
        of the differences, standard deviation of the differences, and percentiles of the absolute differences
        for 68%, 95%, and 99%.
    """
    if eof_number != 1 and eof_number != 2:
        raise ValueError("Argument eof_number must be 1 or 2.")
    doys = tools.doy_list()
    if exclude_doy366:
        doys = doys[:-1]
    corr = np.empty(doys.size)
    diff_mean = np.empty(doys.size)
    diff_std = np.empty(doys.size)
    diff_abs_percent68 = np.empty(doys.size)
    diff_abs_percent95 = np.empty(doys.size)
    diff_abs_percent99 = np.empty(doys.size)

    for idx, doy in enumerate(doys):
        if eof_number == 1:
            eof_ref = eofs_ref.eof1vector_for_doy(doy)
            eof_test = eofs.eof1vector_for_doy(doy)
        elif eof_number == 2:
            eof_ref = eofs_ref.eof2vector_for_doy(doy)
            eof_test = eofs.eof2vector_for_doy(doy)
        (corr_single, diff_mean_single, diff_std_single, diff_vec_single, diff_abs_percent68_single,
         diff_abs_percent95_single, diff_abs_percent99_single) \
            = calc_vector_agreement(eof_ref, eof_test, percentage=percentage, do_print=False)
        corr[idx] = corr_single
        diff_mean[idx] = diff_mean_single
        diff_std[idx] = diff_std_single
        diff_abs_percent68[idx] = diff_abs_percent68_single
        diff_abs_percent95[idx] = diff_abs_percent95_single
        diff_abs_percent99[idx] = diff_abs_percent99_single

    if do_print:
        print("########## Summary of EOF comparison for all DOYs (EOF %i)" % eof_number)
        print("Worst Correlation (at DOY %i): %1.4f" % (doys[np.argmin(corr)], np.amin(corr)))
        print("Worst 99%% percentile (at DOY %i): %1.4f" % (doys[np.argmax(diff_abs_percent99)],
                                                            np.amax(diff_abs_percent99)))
        print("Worst 68%% percentile (at DOY %i): %1.4f" % (doys[np.argmax(diff_abs_percent68)],
                                                            np.amax(diff_abs_percent68)))

    return corr, diff_mean, diff_std, diff_abs_percent68, diff_abs_percent95, diff_abs_percent99


def plot_comparison_stats_for_eofs_all_doys(recalc_eof: eof.EOFDataForAllDOYs,
                                            orig_eof: eof.EOFDataForAllDOYs,
                                            exclude_doy366: bool = False,
                                            do_print: bool = False) -> Figure:
    """
    Plots extended comparison statistics between calculated EOFs and references EOFs.

    Statistics of the differences will be shown for each DOY (DOY on the abscissa) and with one line each for EOF1 and
    EOF2.

    :param recalc_eof: The EOFs to compare.
    :param orig_eof: The reference eofs.
    :param exclude_doy366: If ``True``, DOY 366 will be included in the plot (sometimes worse agreement depending on the
        leap year treatment mode).
    :param do_print: If ``True``, some characteristic values will we printed to the console.

    :return: The figure handle
    """

    doys = tools.doy_list()
    if exclude_doy366:
        doys = doys[:-1]
    xlim = (0, doys[-1])
    (corr_1, diff_mean_1, diff_std_1, diff_abs_percent68_1, diff_abs_percent95_1, diff_abs_percent99_1)\
        = calc_comparison_stats_for_eofs_all_doys(orig_eof, recalc_eof, exclude_doy366=exclude_doy366, eof_number=1,
                                                  percentage=False, do_print=do_print)
    (corr_2, diff_mean_2, diff_std_2, diff_abs_percent68_2, diff_abs_percent95_2, diff_abs_percent99_2)\
        = calc_comparison_stats_for_eofs_all_doys(orig_eof, recalc_eof, exclude_doy366=exclude_doy366, eof_number=2,
                                                  percentage=False, do_print=do_print)

    fig_id = "plot_comparison_stats_for_eofs_all_doys"
    fig, axs = plt.subplots(4, 1, num=fig_id, clear=True, figsize=(9, 9), dpi=150)
    plt.subplots_adjust(hspace=0.5)

    fig.suptitle("Comparison of EOFs for all DOYs")

    ax = axs[0]
    ax.set_title("Correlation")
    p11, = ax.plot(doys, corr_1, label="EOF1", color="blue")
    p12, = ax.plot(doys, corr_2, label="EOF2", color="green")
    ax.set_xlim(xlim)
    ax.set_ylabel("Correlation")
    fig.legend(handles=[p11, p12])

    ax = axs[1]
    ax.set_title("Mean of differences of EOF vector elements")
    p21, = ax.plot(doys, diff_mean_1, label="EOF1", color="blue")
    p22, = ax.plot(doys, diff_mean_2, label="EOF2", color="green")
    ax.set_xlim(xlim)
    ax.set_ylabel(r"Mean [$\mathrm{W/m^2}$]")

    ax = axs[2]
    ax.set_title("Standard deviation of differences of EOF vector elements")
    p31, = ax.plot(doys, diff_std_1, label="EOF1", color="blue")
    p32, = ax.plot(doys, diff_std_2, label="EOF2", color="green")
    ax.set_xlim(xlim)
    ax.set_ylabel(r"Std.Dev. [$\mathrm{W/m^2}$]")

    ax = axs[3]
    ax.set_title("Percentiles of absolute differences of EOF vector elements")
    p31, = ax.plot(doys, diff_abs_percent99_1, label="99% EOF1", color="blue")
    p32, = ax.plot(doys, diff_abs_percent99_2, label="99% EOF2", color="green")
    p33, = ax.plot(doys, diff_abs_percent95_1, label="EOF1", color="blue", linestyle="--")
    p34, = ax.plot(doys, diff_abs_percent95_2, label="EOF2", color="green", linestyle="--")
    p35, = ax.plot(doys, diff_abs_percent68_1, label="EOF1", color="blue", linestyle=":")
    p36, = ax.plot(doys, diff_abs_percent68_2, label="EOF2", color="green", linestyle=":")
    ax.set_xlim(xlim)
    ax.set_ylabel(r"Percentiles [$\mathrm{W/m^2}$]")
    ax.legend(labels=["99%", "95%", "68%"], handles=[p31, p33, p35], loc="upper right")
    ax.set_xlabel("Day of year")

    return fig


def plot_correlation_for_eofs_all_doys(recalc_eof: eof.EOFDataForAllDOYs,
                                       orig_eof: eof.EOFDataForAllDOYs,
                                       exclude_doy366: bool = False,
                                       do_print: bool = False,
                                       full_value_range: bool = True) -> Figure:
    """
    Plots the correlations between calculated EOFs and reference EOFs.

    Correlations will be shown for each DOY (DOY on the abscissa) and with one line each for EOF1 and EOF2.

    :param recalc_eof: The EOFs to validate.
    :param orig_eof: The reference EOFs.
    :param exclude_doy366: If ``False``, DOY 366 will be included in the plot (sometimes worse correlation depending on the
        leap year treatment mode).
    :param do_print: If ``True``, some characteristic values will we printed to the console.
    :param full_value_range: If ``True``, the ordinate spans the range from 0 to 1 instead of the used value range only.
    :return: A handle to the figure.
    """

    doys = tools.doy_list()
    if exclude_doy366:
        doys = doys[:-1]
    xlim = (0, doys[-1])
    (corr_1, diff_mean_1, diff_std_1, diff_abs_percent68_1, diff_abs_percent95_1, diff_abs_percent99_1)\
        = calc_comparison_stats_for_eofs_all_doys(orig_eof, recalc_eof, exclude_doy366=exclude_doy366, eof_number=1,
                                                  percentage=False, do_print=do_print)
    (corr_2, diff_mean_2, diff_std_2, diff_abs_percent68_2, diff_abs_percent95_2, diff_abs_percent99_2)\
        = calc_comparison_stats_for_eofs_all_doys(orig_eof, recalc_eof, exclude_doy366=exclude_doy366, eof_number=2,
                                                  percentage=False, do_print=do_print)

    fig_id = "plot_correlation_for_eofs_all_doys"
    fig, axs = plt.subplots(1, 1, num=fig_id, clear=True, figsize=(6, 4.5), dpi=150)
    plt.subplots_adjust(hspace=0.3)

    fig.suptitle("Correlation of EOFs for all DOYs")

    ax = axs
    p11, = ax.plot(doys, corr_1, label="EOF1", color="blue")
    p12, = ax.plot(doys, corr_2, label="EOF2", color="green")
    ax.set_xlim(xlim)
    ax.legend(handles=[p11, p12])
    ax.set_xlabel("Day of year")
    ax.set_ylabel("Correlation coefficient")
    if full_value_range:
        ax.set_ylim((0, 1.1))

    return fig


def plot_individual_eof_map_comparison(orig_eof: eof.EOFData, compare_eof: eof.EOFData, doy: int = None) -> Figure:
    """
    Shows the maps of EOFs 1 and 2 and a respective reference together with a map of differences between both.

    :param orig_eof: The reference EOF data.
    :param compare_eof: The EOF data to validate.
    :param doy: The DOY, which is evaluated (only used for the figure title).

    :return: The figure handle.
    """

    fig, axs = plt.subplots(2, 3, num="ReproduceOriginalOMIPCs_ExplainedVariance_EOF_Comparison", clear=True,
                            figsize=(10, 5), dpi=150, sharex=True, sharey=True)
    plt.subplots_adjust(wspace=0.35, hspace=0.35)
    if doy is not None:
        fig.suptitle("EOF Recalculation for DOY %i" % doy)

    ax = axs[0, 0]
    c = ax.contourf(orig_eof.long, orig_eof.lat, orig_eof.eof1map,
                    levels=np.arange(-0.1, 0.11, 0.01), cmap=matplotlib.cm.get_cmap("bwr"))
    fig.colorbar(c, ax=ax, label="OLR Anomaly [W/m²]")
    ax.set_title("Original EOF1")
    ax.set_ylabel("Latitude [°]")

    ax = axs[0, 1]
    c = ax.contourf(compare_eof.long, compare_eof.lat, compare_eof.eof1map,
                    levels=np.arange(-0.1, 0.11, 0.01), cmap=matplotlib.cm.get_cmap("bwr"))
    fig.colorbar(c, ax=ax, label="OLR Anomaly [W/m²]")
    ax.set_title("Recalculated EOF1")

    # FIXME: Check that grids are equal
    ax = axs[0, 2]
    c = ax.contourf(orig_eof.long, orig_eof.lat, orig_eof.eof1map - compare_eof.eof1map,
                    levels=100, cmap=matplotlib.cm.get_cmap("Purples"))
    fig.colorbar(c, ax=ax, label="OLR Anomaly [W/m²]")
    ax.set_title("Difference 1")

    ax = axs[1, 0]
    c = ax.contourf(orig_eof.long, orig_eof.lat, orig_eof.eof2map,
                    levels=np.arange(-0.1, 0.11, 0.01), cmap=matplotlib.cm.get_cmap("bwr"))
    fig.colorbar(c, ax=ax, label="OLR Anomaly [W/m²]")
    ax.set_title("Original EOF2")
    ax.set_ylabel("Latitude [°]")
    ax.set_xlabel("Longitude [°]")

    ax = axs[1, 1]
    c = ax.contourf(compare_eof.long, compare_eof.lat, compare_eof.eof2map,
                    levels=np.arange(-0.1, 0.11, 0.01), cmap=matplotlib.cm.get_cmap("bwr"))
    fig.colorbar(c, ax=ax, label="OLR Anomaly [W/m²]")
    ax.set_title("Recalculated EOF2")
    ax.set_xlabel("Longitude [°]")

    ax = axs[1, 2]
    c = ax.contourf(orig_eof.long, orig_eof.lat, orig_eof.eof2map - compare_eof.eof2map,
                    levels=100, cmap=matplotlib.cm.get_cmap("Purples"))
    fig.colorbar(c, ax=ax, label="OLR Anomaly [W/m²]")
    ax.set_title("Difference 2")
    ax.set_xlabel("Longitude [°]")

    return fig


def calc_comparison_stats_for_explained_variance(ref_var: np.ndarray,
                                                 calc_var: np.ndarray,
                                                 do_print: bool = False,
                                                 exclude_doy366: bool = False) -> typing.Tuple:
    """
    Calculates the comparison statistics of the explained variances for one EOF and all DOYs.

    :param ref_var: The reference variances.
    :param calc_var: The variances to compare.
    :param do_print: If ``True``, some statistical values will also be shown in the console.
    :param exclude_doy366: If ``True``, the data for DOY 366 will not be considered in the statistics.

    :return: A tuple containing values for the following quantities: correlation, mean of the differences,
        standard deviation of the differences, and percentiles of the absolute differences for 68%, 95%, and 99%.
    """
    ref_data = ref_var.copy()
    calc_data = calc_var.copy()
    if exclude_doy366:
        if do_print:
            print("###### DOY 366 excluded")
        ref_data = ref_data[:-1]
        calc_data = calc_data[:-1]
    (corr, diff_mean, diff_std, diff_vec, diff_abs_percent68, diff_abs_percent95, diff_abs_percent99)\
        = calc_vector_agreement(ref_data, calc_data, percentage=False, do_print=do_print)
    return corr, diff_mean, diff_std, diff_vec, diff_abs_percent68, diff_abs_percent95, diff_abs_percent99


def plot_comparison_stats_for_explained_variance(ref_var: np.ndarray,
                                                 calc_var: np.ndarray,
                                                 title: str = None,
                                                 do_print: bool = False,
                                                 exclude_doy366: bool = False) -> Figure:
    """
    Plots the comparison of the explained variances for one EOF and for all DOYs.

    :param ref_var: The reference variances.
    :param calc_var: The variances to validate.
    :param title: A title for the figure.
    :param do_print: If ``True``, some statistical values will also be shown in the console.
    :param exclude_doy366: If ``True``, the data for DOY 366 will not be considered in the statistics.

    :return: The figure handle
    """
    if do_print:
        print("##########")
        if title is not None:
            print(title)
    ref_data = ref_var.copy()
    calc_data = calc_var.copy()
    if exclude_doy366:
        if do_print:
            print("###### DOY 366 excluded")
        ref_data = ref_data[:-1]
        calc_data = calc_data[:-1]
    fig = plot_vector_agreement(ref_data, calc_data, title=title, do_print=do_print)
    return fig


def calc_timeseries_agreement(ref_data: np.ndarray, ref_time: np.ndarray, data: np.ndarray, time: np.ndarray,
                              exclude_doy366: bool = False, do_print: bool = False) -> typing.Tuple:
    """
    Calculates comparison values of two time series.

    :param ref_data: The reference time series vector.
    :param ref_time: The time grid of the reference.
    :param data: The time series vector to validate.
    :param time: The time grid of the time series to validate. It will be checked if this is similar to the time grid
        of the reference.
    :param exclude_doy366: If ``True``, the data for DOY 366 will not be considered in the statistics.
    :param do_print: If ``True``, some statistical values will also be shown in the console.

    :return: A tuple containing values for the following quantities: correlation, mean of the differences,
        standard deviation of the differences, and percentiles of the absolute differences for 68%, 95%, and 99%.
    """
    if not np.all(ref_time == time):
        raise ValueError("Time series do not cover the same periods.")
    if exclude_doy366:
        if do_print:
            print("###### DOY 366 excluded")
        doys = tools.calc_day_of_year(ref_time)
        inds_used = np.nonzero(doys != 366)
        inds_not_used = np.nonzero(doys == 366)
        calc_ref_data = ref_data[inds_used]
        calc_data = data[inds_used]
    else:
        if do_print:
            print("##### Complete time series")
        calc_ref_data = ref_data
        calc_data = data
        inds_used = (np.arange(0, ref_time.size, 1),)
        inds_not_used = (np.array([], dtype="int64"),)

    (corr, diff_mean, diff_std, diff_ts, diff_abs_percent68, diff_abs_percent95, diff_abs_percent99)\
        = calc_vector_agreement(calc_ref_data, calc_data, percentage=False, do_print=do_print)

    return (inds_used, inds_not_used, corr, diff_mean, diff_std, diff_ts, diff_abs_percent68, diff_abs_percent95,
            diff_abs_percent99)


def plot_timeseries_agreement(ref_data: np.ndarray,
                              ref_time: np.ndarray,
                              data: np.ndarray,
                              time: np.ndarray,
                              title: str = None,
                              do_print: bool = False) -> Figure:
    """
    Plots a graphical comparison of 2 time series.

    Shows 4 subplots with 1) the data, 2) the difference, 3) a scatterplot of the data, and 4) a histogram of the
    differences.

    :param ref_data: The time series reference vector.
    :param ref_time: The time grid of the reference.
    :param data: The time series vector to validate.
    :param time: The time grid of the data to validate. It will be checked if this is similar to the time grid of the reference.
    :param title: A title for the plot. Use this to explain the quantity which is compared evaluated with the plot.
    :param do_print: If ``True``, some statistical values will also be shown in the console.

    :return: The figure handle.
    """
    if do_print:
        print("##########")
        if title is not None:
            print(title)

    if not (time[0] == ref_time[0]) or not (time[-1] == ref_time[-1]):
        (time_intersect, data_ind, ref_data_ind) = np.intersect1d(time, ref_time, return_indices=True)
        if time_intersect.size == 0:
            raise ValueError("Provided time series to plot do not overlap.")
        if not np.all(np.diff(time_intersect) == np.diff(time_intersect)[0]):
            raise ValueError("Combined time grid of both time series to plot is not equidistant.")
        time = time_intersect
        data = data[data_ind]
        ref_time = time_intersect
        ref_data = ref_data[ref_data_ind]
        message = ("The provided time series do not cover the same period.\nThe compared range was restricted to the "
                   "overlapping period from {0} to {1}.\nStronger differences may occur at least at the end of the "
                   "compared period, since the temporal filtering worked on different samples."
                   .format(str(time[0]), str(time[-1])))
        warnings.warn(message)

    (tempa, tempb, corr_complete, diff_mean_complete, diff_std_complete, diff_ts_abs_complete,
     diff_abs_percent68_complete, diff_abs_percent95_complete, diff_abs_percent99_complete) \
        = calc_timeseries_agreement(ref_data, ref_time, data, time, exclude_doy366=False, do_print=do_print)

    (doy_not_366_inds, doy366_inds, corr_not366, diff_mean_not366, diff_std_not366, diff_ts_abs_not366,
     diff_abs_percent68_not366, diff_abs_percent95_not366, diff_abs_percent99_not366) \
        = calc_timeseries_agreement(ref_data, ref_time, data, time, exclude_doy366=True, do_print=do_print)

    ref_data_not366 = ref_data[doy_not_366_inds]
    data_not366 = data[doy_not_366_inds]

    fig_id = "evaluate_timeseries_agreement"
    if title is not None:
        fig_id = fig_id + "_" + title
    fig, axs = plt.subplots(2, 2, num=fig_id, clear=True, figsize=(12, 9), dpi=150)
    plt.subplots_adjust(hspace=0.3)
    if title is not None:
        fig.suptitle(title)

    ax = axs[0, 0]
    ax.set_title("Time series")
    p11, = ax.plot(time, ref_data, label="Reference")
    p12, = ax.plot(time, data, label="Recalculation")
    ax.set_xlabel("Time")
    ax.set_ylabel("Data [units of quantity]")

    ax = axs[1, 0]
    ax.set_title("Difference")
    p21, = ax.plot(time, diff_ts_abs_complete, color="red")
    p22, = ax.plot(time[doy_not_366_inds], diff_ts_abs_not366, color="blue")
    ax.set_xlabel("Time")
    ax.set_ylabel("Difference [units of quantity]")

    ax = axs[0, 1]
    ax.set_title("Scatterplot")
    p31, = ax.plot(ref_data_not366, data_not366, linestyle='None', marker="x", color="blue")
    p32, = ax.plot(ref_data[doy366_inds], data[doy366_inds], linestyle='None', marker="x", color="red")
    ax.set_xlabel("Reference data [units of quantity]")
    ax.set_ylabel("Data [units of quantity]")

    ax = axs[1, 1]
    (n2, bins2, patches) = ax.hist(diff_ts_abs_not366, bins=100, density=True)
    ax.set_xlabel("Difference [units of quantity]")
    ax.set_ylabel("Number of occurrences")
    sigma = diff_std_not366
    mu = diff_mean_not366
    ax.plot(bins2, 1 / (sigma * np.sqrt(2 * np.pi)) * np.exp(- (bins2 - mu) ** 2 / (2 * sigma ** 2)),
            linewidth=2, color='r')

    return fig


def plot_vector_agreement(ref_data: np.ndarray, data: np.ndarray, title: str = None, do_print: bool = False) -> Figure:
    """
    Plot a graphical comparison of 2 vectors.

    Shows 4 subplots with 1) the data, 2) the difference, 3) a scatterplot of the data, and 4) a histogram of
    the differences.

    :param ref_data: The reference vector.
    :param data: The vector to validate.
    :param title: A title for the plot. Use this to explain the quantity which is evaluated with the plot.
    :param do_print: If ``True``, some statistics values will also be shown in the console.

    :return: The figure handle.
    """
    if do_print:
        print("##########")
        if title is not None:
            print(title)
    corr, diff_mean, diff_std, diff_vec, diff_abs_percent68, diff_abs_percent95, diff_abs_percent99 = \
        calc_vector_agreement(ref_data, data, percentage=False, do_print=do_print)

    fig_id = "evaluate_vector_agreement"
    if title is not None:
        fig_id = fig_id + "_" + title
    fig, axs = plt.subplots(2, 2, num=fig_id, clear=True, figsize=(12, 9), dpi=150)
    plt.subplots_adjust(hspace=0.3)
    if title is not None:
        fig.suptitle(title)

    ax = axs[0, 0]
    ax.set_title("Data")
    p11, = ax.plot(ref_data, label="Reference")
    p12, = ax.plot(data, label="Recalculation")
    ax.set_xlabel("Position in vector")
    ax.set_ylabel("Data [units of quantity]")

    ax = axs[1, 0]
    ax.set_title("Difference")
    p21, = ax.plot(diff_vec, color="red")
    ax.set_xlabel("Position in vector")
    ax.set_ylabel("Difference [units of quantity]")

    ax = axs[0, 1]
    ax.set_title("Scatterplot")
    p31, = ax.plot(ref_data, data, linestyle='None', marker="x", color="blue")
    ax.set_xlabel("Reference data [units of quantity]")
    ax.set_ylabel("Data [units of quantity]")

    ax = axs[1, 1]
    (n2, bins2, patches) = ax.hist(diff_vec, bins=100, density=True)
    ax.set_xlabel("Difference [units of quantity]")
    ax.set_ylabel("Number of occurrences")
    sigma = diff_std
    mu = diff_mean
    ax.plot(bins2, 1 / (sigma * np.sqrt(2 * np.pi)) * np.exp(- (bins2 - mu) ** 2 / (2 * sigma ** 2)),
            linewidth=2, color='r')

    return fig


def plot_comparison_orig_calc_pcs(calc_pcs: pc.PCData,
                                  orig_pcs: pc.PCData,
                                  start_date: np.datetime64 = None,
                                  end_date: np.datetime64 = None):
    """
    Plots both PC time series (one in a subplot each) of the recalculation and a reference.

    The period to plot can be adjusted.

    :param calc_pcs: The recalculated PC time series.
    :param orig_pcs: The reference PC time series.
    :param start_date: Start of the period to plot. If ``None``, the whole period will be plotted.
    :param end_date: End of the period to plot. If ``None``, the whole period will be plotted.

    :return: The figure handle.
    """
    fig, axs = plt.subplots(2, 1, num="plot_comparison_orig_calc_pcs", clear=True, figsize=(8, 6), dpi=150)
    plt.subplots_adjust(hspace=0.4)
    fig.suptitle("PC Recalculation")

    ax = axs[0]
    ax.set_title("Principal Component 1")
    p1, = ax.plot(orig_pcs.time, orig_pcs.pc1, label="Original")
    p2, = ax.plot(calc_pcs.time, calc_pcs.pc1, label="Recalculation")
    if start_date is not None and end_date is not None:
        ax.set_xlim((start_date, end_date))
    ax.set_ylabel("PC1 [1]")
    plt.setp(ax.get_xticklabels(), rotation=15, horizontalalignment='right')
    fig.legend(handles=(p1, p2))

    ax = axs[1]
    ax.set_title("Principal Component 2")
    p3, = ax.plot(orig_pcs.time, orig_pcs.pc2, label="Original")
    p4, = ax.plot(calc_pcs.time, calc_pcs.pc2, label="Recalculation")
    if start_date is not None and end_date is not None:
        ax.set_xlim((start_date, end_date))
    ax.set_ylabel("PC2 [1]")
    plt.setp(ax.get_xticklabels(), rotation=15, horizontalalignment="right")

    return fig


def _explained_variance_file_converter(s: str) -> float:
    return float(re.sub(r"[ \[;\]]", "", s.decode("utf-8")))


def load_omi_explained_variance(filename: str) -> typing.Tuple:
    """
    Loads original explained variance files provided by Juliana Dias.

    :param filename: The file to load.

    :return: A tuple containing two arrays, for EOF1 and EOF2, respectively.
    """
    data = np.genfromtxt(filename,
                         converters={0: _explained_variance_file_converter, 1: _explained_variance_file_converter},
                         skip_header=7)
    var1 = data[:, 0]
    var2 = data[:, 1]
    return var1, var2



# -*- coding: utf-8 -*-

# Copyright (C) 2019 Christoph G. Hoffmann. All rights reserved.

# This file is part of mjoindices

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Contact: christoph.hoffmann@uni-greifswald.de

"""
This module provides the 2-dim (temporal and longitudinal) filtering algorithm used by the OMI approach.

Although implemented generically, this module is not intended to be used stand-alone outside the OMI context,
as it has only been extensively tested for the specific OMI filtering conditions.

Hence, there is usually no need for the users of the mjoindices package to call functions of this module themselves.
Instead, they probably want to use the module :py:mod:`mjoindices.omi.omi_calculator` directly.

The complete algorithm is described by :ref:`refKiladis2014`
"""

import matplotlib.pyplot as plt
import numpy as np
import mjoindices.olr_handling as olr


def filter_olr_for_mjo_pc_calculation(olrdata: olr.OLRData, do_plot: bool = False):
    """
    Filters OLR data temporally with a bandwidth particularly selected for the PC calculation.

    The filter algorithm is the same as for the combined temporal and longitudinal filtering,
    but the longitudinal bandpass filter constants are defined so broadly that effectively no longitudinal filtering is
    applied. The temporal filtering constants are chosen to meet the values in the description by :ref:`refKiladis2014`.

    :param olrdata: The original OLR data.
    :param do_plot: If ``True``, diagnosis plots will be generated.

    :return: The filtered OLR.
    """
    return filter_olr_temporally(olrdata, 20., 96., do_plot=do_plot)


# Implicitly tested for special conditions with specific caller functions
def filter_olr_temporally(olrdata: olr.OLRData, period_min: float, period_max: float, do_plot: bool = False):
    """
    Filters OLR data temporally.

    The filter algorithm is the same as for the combined temporal and longitudinal filtering,
    but the longitudinal bandpass filter constants are defined so broadly that effectively no longitudinal filtering is
    applied.

    Note that this function has only been strictly tested for filtering constants used by the OMI algorithm.

    :param olrdata: The original OLR data
    :param period_min: Temporal filter constant: Only greater periods (in days) remain in the data.
    :param period_max: Temporal filter constant: Only lower periods (in days) remain in the data.
    :param do_plot: If ``True``, diagnosis plots will be generated.

    :return: The filtered OLR.
    """
    return filter_olr_temporally_and_longitudinally(olrdata, period_min, period_max, -720., 720, do_plot=do_plot)


def filter_olr_for_mjo_eof_calculation(olrdata: olr.OLRData, do_plot: bool = False) -> olr.OLRData:
    """
    Filters OLR data temporally and longitudinally with a bandwidth particularly selected for the EOF calculation.

    The filter setup meets the description of Kiladis (2014) for the EOF calculation.

    :param olrdata: The original OLR data
    :param do_plot: If ``True``, diagnosis plots will be generated.

    :return: The filtered OLR data.
    """
    return filter_olr_temporally_and_longitudinally(olrdata, 30., 96., 0., 720, do_plot=do_plot)


# Implicitly tested for special conditions with specific caller functions
def filter_olr_temporally_and_longitudinally(olrdata: olr.OLRData,
                                             period_min: float,
                                             period_max: float,
                                             wn_min: float,
                                             wn_max: float,
                                             do_plot: bool=False) -> olr.OLRData:
    """
    Performs a temporal and longitudinal bandpass filtering of the OLR data with configurable filtering thresholds.

    Note that this function has only been strictly tested for filtering constants used by the OMI algorithm.

    :param olrdata: The original OLR data.
    :param period_min: Temporal filter constant: Only greater periods (in days) remain in the data.
    :param period_max: Temporal filter constant: Only lower periods (in days) remain in the data.
    :param wn_min: Longitudinal filter constant: Only greater wave numbers (in cycles per globe) remain in the data.
    :param wn_max:  Longitudinal filter constant: Only lower wave numbers (in cycles per globe) remain in the data.
    :param do_plot: If ``True``, diagnosis plots will be generated.

    :return: The filtered OLR.
    """
    print("Smooth data temporally and longitudinally...")
    filtered_olr = np.empty(olrdata.olr.shape)

    for ilat, lat in enumerate(olrdata.lat):
        print("Filtering for latitude: ", lat)
        time_spacing = (olrdata.time[1] - olrdata.time[0]).astype('timedelta64[s]') / np.timedelta64(1, 'D')  # time spacing in days
        dataslice = np.squeeze(olrdata.olr[:, ilat, :])
        wkfilter = WKFilter()
        filtered_data = wkfilter.perform_2dim_spectral_filtering(dataslice, time_spacing, period_min, period_max, wn_min,
                                                                 wn_max, do_plot=do_plot, save_debug=False)
        filtered_olr[:, ilat, :] = filtered_data

    return olr.OLRData(filtered_olr, olrdata.time, olrdata.lat, olrdata.long)


def detrend_vector(data: np.ndarray) -> np.ndarray:
    """
    Removes the trend from the given vector.

    :param data: The vector to detrend

    :return: The data with removed trend.
    """
    x = np.arange(0, data.size, 1)
    A = np.vstack([x, np.ones(len(x))]).T
    m, b = np.linalg.lstsq(A, data, rcond=-1)[0]

    result = data - (m * x + b)
    #    plt.plot(ts)
    #    plt.plot(m*x+b)
    #    plt.plot(result)
    return result


def taper_vector_to_zero(data: np.ndarray, window_length: int) -> np.ndarray:
    """
    Taper the data in the given vector to zero at both the beginning and the ending.

    :param data: The data to taper.
    :param window_length: The length of the window (measured in vector indices),
        in which the tapering is applied for the beginning and the ending independently

    :return: The tapered data.
    """
    startinds = np.arange(0, window_length, 1)
    endinds = np.arange(-window_length - 1, -1, 1) + 2

    result = data
    result[0:window_length] = result[0:window_length] * 0.5 * (1 - np.cos(startinds * np.pi / window_length))
    result[data.size - window_length:data.size] = \
        result[data.size - window_length:data.size] * 0.5 * (1 - np.cos(endinds * np.pi / window_length))
    return result


class WKFilter:
    """
    This class contains the major Wheeler-Kiladis-Filtering functionality.
    The functionality is encapsulated in a class because values of intermediate processing steps
    are saved as class members for debugging purposes.
    To run the filtering, only the method :py:func:`perform_2dim_spectral_filtering` has to be executed.
    """
    def __init__(self):
        self.DebugInputOLR = []
        self.DebugFilterOLR = []
        self.DebugDetrendedOLR = []
        self.DebugPreprocessedOLR = []
        self.DebugFreqAxis = []
        self.DebugWNAxis = []
        self.DebugOriginalFourierSpectrum = []
        self.DebugFilteredFourierSpectrum = []
        self.DebugNoElementsInFilteredSpectrum = []

    def perform_2dim_spectral_filtering(self,
                                        data: np.ndarray,
                                        time_spacing: float,
                                        period_min: float,
                                        period_max: float,
                                        wn_min: float,
                                        wn_max: float,
                                        do_plot: bool = False,
                                        save_debug: bool = False) -> np.ndarray:
        """
        Bandpass-filters OLR data in time- and longitude-direction according to
        the original Kiladis algorithm.

        Note that the temporal and longitudinal dimension have in principle
        different characteristics, so that they are in detail treated a bit
        differently:
        While the time variable may evolve to infinity (thus, the number of data
        points and the time_spacing variable are needed to calculate the
        full temporal coverage), longitude is periodic with the periodicity
        of one globe (thus, it is assumed that the data covers exactly one
        globe and only passing the number of longitudes provides complete information).

        :param data: The OLR data as 2-dim array: first dimension time, second
            dimension longitude, both equally spaced.
            The longitudinal dimension has to cover the full globe.
            The time dimension is further described by the variable
            `time_spacing`.
        :param time_spacing: Temporal resolution of the data in days (often 1 or 0.5 (if two
            data points exist per day)).
        :param period_min: Minimal period (in days) that remains in the dataset.
        :param period_max: Maximal period (in days) that remains in the dataset.
        :param wn_min: Minimal wavenumber (in cycles per globe) that remains in the dataset.
        :param wn_max: Maximal wavenumber (in cycles per globe) that remains in the dataset.
        :param do_plot: If ``True``, diagnosis plots will be generated.
        :param save_debug: If ``True``, some variables will be filled with values of intermediate processing steps
            for debugging purposes.

        :return: The filtered data.
        """

        # ###################### Process input data #######################
        if save_debug:
            self.DebugInputOLR = np.copy(data)

        if do_plot:
            fig = plt.figure("WK_Filter_perform2dimSpectralSmoothing_DataIn", clear=True)
            plt.contourf(data)
            plt.colorbar()
            plt.title("Original Data")

        dataperday = 1 / time_spacing
        freq_min = 1 / period_max
        freq_max = 1 / period_min

        # ######################## Detrend #################################
        # "orig" refers to the original size in the time dimension in the following, i.e. not the zero-padded version.
        orig_data = data
        orig_nt, nl = orig_data.shape

        for idx_l in range(0, nl):
            orig_data[:, idx_l] = detrend_vector(orig_data[:, idx_l])

        if save_debug:
            self.DebugDetrendedOLR = np.copy(orig_data)
        if do_plot:
            fig = plt.figure("WK_Filter_perform2dimSpectralSmoothing_Detrended", clear=True)
            plt.contourf(orig_data)
            plt.colorbar()
            plt.title("Detrended Data")

        # ######################## Zero Padding ############################
        nt = 2 ** 17  # Zero padding for performance and resolution optimization, as well as consistency with origininal Kiladis code

        if orig_nt > nt:
            raise ValueError('Time series is longer than hard-coded value for zero-padding!')

        data = np.zeros([nt, nl])
        data[0:orig_nt, :] = orig_data

        # ######################## Tapering to zero ########################
        # 10 days tapering according ot Kiladis Code
        # only relevant at beginning of time series as it is zero-padded in the end
        for idx_l in range(0, nl):
            data[:, idx_l] = taper_vector_to_zero(data[:, idx_l], int(10 * dataperday))

        if save_debug:
            self.DebugPreprocessedOLR = np.copy(data)

        # ########################## Forward Fourier transform ############
        fourier_fft = np.fft.fft2(data)
        # reordering of spectrum is done to be consistent with the original kiladis ordering.
        fourier_fft = np.fft.fftshift(fourier_fft, axes=(0, 1))
        fourier_fft = np.roll(fourier_fft, int(nt / 2), axis=0)
        fourier_fft = np.roll(fourier_fft, int(nl / 2), axis=1)

        # ## Calculation of the frequency grid in accordance with Kiladis code
        freq_axis = np.zeros(nt)
        for i_f in range(0, nt):
            if (i_f <= nt / 2):
                freq_axis[i_f] = i_f * dataperday / nt
            else:
                freq_axis[i_f] = -1 * (nt - i_f) * dataperday / nt
        # the following code based on scipy function produces qualitatively the same grid.
        # However, numerical differences seem to have larger effect for the filtering step.
        # freq_axis = np.fft.fftfreq(nt, d=time_spacing)
        # freq_axis = np.fft.fftshift(freq_axis)
        # freq_axis = np.roll(freq_axis, int(nt/2))

        # ## Calculation of the wavenumber grid in accordance with Kiladis code
        wn_axis = np.zeros(nl)
        for i_wn in range(0, nl):
            if i_wn <= nl / 2:
                wn_axis[i_wn] = -1 * i_wn
                # note: to have this consistent with the time-dimension, one could write wn_axis[i_wn]= -1*i_wn*dataperglobe/nl
                # However, since data is required to cover always one globe nl will always be equal to dataperglobe
                # The sign is not consistent with the time dimension, which is for reasons of consitency with the original Kiladis implementation
            else:
                wn_axis[i_wn] = nl - i_wn
        # the following code based on scipy function produces qualitatively the same grid.
        # However, numerical differences seem to have larger effect for the filtering step.
        # wn_axis = np.fft.fftfreq(nl, d=dy)
        # wn_axis = np.fft.fftshift(wn_axis)  #identical with  wn_axis=np.arange(-int(nlong/2), int(nlong/2),1.)
        # wn_axis = -1 *wn_axis
        # wn_axis = np.roll(wn_axis, int(nl/2))

        if save_debug:
            self.DebugFreqAxis = np.copy(freq_axis)
            self.DebugWNAxis = np.copy(wn_axis)
            self.DebugOriginalFourierSpectrum = np.copy(fourier_fft)

        if do_plot:
            fig = plt.figure("WK_Filter_perform2dimSpectralSmoothing_freqAxis", clear=True)
            plt.plot(freq_axis)
            plt.title("Calc freq axis")

            fig = plt.figure("WK_Filter_perform2dimSpectralSmoothing_WNAxis", clear=True)
            plt.plot(wn_axis)
            plt.title("Calc wn axis")

            fig = plt.figure("WK_Filter_perform2dimSpectralSmoothing_Spectrum", clear=True)
            plt.contourf(wn_axis, freq_axis, np.squeeze(fourier_fft))
            plt.colorbar()
            plt.title("Fourier Transformation")

        # ################### Filtering of the Fourier Spectrum #############
        # ## name filter boundaries like in Kiladis Fortran Code
        f1 = freq_min
        f2 = freq_min
        f3 = freq_max
        f4 = freq_max
        s1 = wn_min
        s2 = wn_max
        s3 = wn_min
        s4 = wn_max

        # ### Very similar to original Kiladis Code
        fourier_fft_filtered = fourier_fft
        count = 0
        for i_f in range(0, int(nt / 2) + 1):
            for i_wn in range(0, nl):
                ff = freq_axis[i_f]
                ss = wn_axis[i_wn]
                if ((ff >= ((ss * (f1 - f2) + f2 * s1 - f1 * s2) / (s1 - s2))) and
                        (ff <= ((ss * (f3 - f4) + f4 * s3 - f3 * s4) / (s3 - s4))) and
                        (ss >= ((ff * (s3 - s1) - f1 * s3 + f3 * s1) / (f3 - f1))) and
                        (ss <= ((ff * (s4 - s2) - f2 * s4 + f4 * s2) / (f4 - f2)))):
                    count = count + 1
                else:
                    fourier_fft_filtered[i_f, i_wn] = 0
                    if i_wn == 0 and i_f == 0:
                        pass
                    elif i_wn == 0:
                        ind_f = nt - i_f
                        if (ind_f < nt):
                            fourier_fft_filtered[ind_f, i_wn] = 0
                    elif i_f == 0:
                        ind_wn = nl - i_wn
                        if ind_wn < nl:
                            fourier_fft_filtered[i_f, ind_wn] = 0
                    else:
                        ind_f = nt - i_f
                        ind_wn = nl - i_wn
                        if ind_f < nt and ind_wn < nl:
                            fourier_fft_filtered[ind_f, ind_wn] = 0
        if save_debug:
            self.DebugFilteredFourierSpectrum = np.copy(fourier_fft_filtered)
            self.DebugNoElementsInFilteredSpectrum = count

        if do_plot:
            fig = plt.figure("WK_Filter_perform2dimSpectralSmoothing_FilteredSpectrum", clear=True)
            plt.contourf(wn_axis, freq_axis, np.squeeze(fourier_fft_filtered))
            plt.colorbar()
            plt.title("Filtered Fourier Transformation")
            print("Number of elements in filtered spectrum: ", count)

        # ############################ FFT Backward transformation ############
        # #reorder spectrum back from kiladis ordering to python ordering
        fourier_fft_filtered = np.roll(fourier_fft_filtered, -int(nt / 2), axis=0)
        fourier_fft_filtered = np.roll(fourier_fft_filtered, -int(nl / 2), axis=1)
        fourier_fft_filtered = np.fft.ifftshift(fourier_fft_filtered, axes=(0, 1))
        filtered_olr = np.fft.ifft2(fourier_fft_filtered)
        filtered_olr = np.real(filtered_olr)

        # ############################# remove zero padding elements ##########
        result = filtered_olr[0:orig_nt, :]

        if save_debug:
            self.DebugFilterOLR = np.copy(result)

        if do_plot:
            fig = plt.figure("perform2dimSpectralSmoothing_4", clear=True)
            plt.contourf(result)
            plt.colorbar()
            plt.title("Filtered Data")
        # ToDo: Make sure that result is real
        return result



# -*- coding: utf-8 -*-

# Copyright (C) 2019 Christoph G. Hoffmann. All rights reserved.

# This file is part of mjoindices

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Contact: christoph.hoffmann@uni-greifswald.de

"""
This major module provides the execution of the algorithm of the OMI calculation.

First, the algorithm can be used to compute the empirical orthogonal functions (EOFs), which serve as a new basis for
the OLR data. According to the OMI algorithm, the EOFs have to be computed for each day of the year (DOY). Second,
the time-dependent coefficients of the OLR maps w.r.t the EOFs are computed. These coefficients are called the
principal components (PCs).

The complete algorithm is described in :ref:`refKiladis2014`.

Basically, only OLR data on a suitable spatial grid is needed. With that, the OMI EOFs and afterwards the PCs are
computed using the functions :func:`calc_eofs_from_olr` and :func:`calculate_pcs_from_olr`, respectively.

In terms of the post-processing of the EOFs (before the PCs are calculated), the following options are available.
They are selected with the parameter ``eofs_postprocessing_type`` of the function :py:func:`calc_eofs_from_olr`

* The original approach is described in :ref:`refKiladis2014`.
  Details are found in :py:func:`~mjoindices.omi.postprocessing_original_kiladis2014.post_process_eofs_original_kiladis_approach`

* An alternative post-processing procedure may be used to reduce noise and potential degeneracy issues. This algorithm is
  described in :ref:`refWeidman2022`. The alternative post-processing procedure is explained
  in :py:func:`~mjoindices.omi.postprocessing_rotation_approach.post_process_eofs_rotation`.

"""

from pathlib import Path
from typing import Tuple

import numpy as np
import warnings
import importlib

import mjoindices.empirical_orthogonal_functions as eof
import mjoindices.olr_handling as olr
import mjoindices.principal_components as pc
import mjoindices.omi.wheeler_kiladis_mjo_filter as wkfilter
import mjoindices.omi.quick_temporal_filter as qfilter
import mjoindices.omi.postprocessing_original_kiladis2014 as pp_kil2014
import mjoindices.omi.postprocessing_rotation_approach as pp_rotation
import mjoindices.tools as tools

eofs_spec = importlib.util.find_spec("eofs")
eofs_package_available = eofs_spec is not None
if eofs_package_available:
    import eofs.standard as eofs_package


# #################EOF calculation

def calc_eofs_from_olr(olrdata: olr.OLRData,
                       implementation: str = "internal",
                       leap_year_treatment: str = "original",
                       eofs_postprocessing_type: str ="kiladis2014",
                       eofs_postprocessing_params: dict=None,
                       sign_doy1reference: bool = None,
                       interpolate_eofs: bool = None,
                       interpolation_start_doy: int = None,
                       interpolation_end_doy: int = None,
                       strict_leap_year_treatment: bool = None
                       ) -> eof.EOFDataForAllDOYs:
    """
    One of the major functions of this module. It performs the complete OMI EOF computation.

    This function executes consistently the preprocessing (filtering), the actual EOF analysis, and the postprocessing.

    :param olrdata: The OLR dataset, from which OMI should be calculated. Note that OLR values are assumed to be given
        in positive values. The spatial grid of the OLR datasets defines also the spatial grid of the complete OMI
        calculation.
    :param implementation: Choose one of the following values:

        * ``"internal"``: uses the internal implementation of the EOF approach.
        * ``"eofs_package"``: Uses the implementation of the external package :py:mod:`eofs`.

    :param leap_year_treatment: Choose one of the following values:

        * ``"original"`` will be as close to the original version of :ref:`refKiladis2014` as possible.
        * ``"strict"`` (not recommended) will treat leap years somewhat more strictly, which might, however,
          cause the results to deviate from the original.
          See also description in :meth:`~mjoindices.tools.find_doy_ranges_in_dates`.
        * ``"no_leap_years"`` will act as if there are no leap years in the dataset (365 days consistently),
          which might be useful for modeled data.

    :param eofs_postprocessing_type: Different approaches of the post-processing of the EOFs are available:

        * ``"kiladis2014"`` for the original post-processing described in :ref:`refKiladis2014`.
        * ``"eof_rotation"`` for the post-processing rotation algorithm described in :ref:`refWeidman2022`;
        * ``None`` for no post-processing.

    :param eofs_postprocessing_params: dict of specific parameters, which will be passed as keyword parameters to the
        respective post-processing function. See the descriptions of the respective functions for details
        (:py:func:`~mjoindices.omi.postprocessing_original_kiladis2014.post_process_eofs_original_kiladis_approach`
        or :py:func:`~mjoindices.omi.postprocessing_rotation_approach.post_process_eofs_rotation`).

    :param sign_doy1reference: .. deprecated:: 1.4
    :param interpolate_eofs: .. deprecated:: 1.4
    :param interpolation_start_doy: .. deprecated:: 1.4
    :param interpolation_end_doy: .. deprecated:: 1.4
    :param strict_leap_year_treatment: .. deprecated:: 1.4

    :return: The computed EOFs.

    """
    # ######The following lines are only for backwards compatibility of the interface and can be deleted in one of the next releases
    deprecated_pp_interface_used = False
    if sign_doy1reference is not None:
        deprecated_pp_interface_used = True
        warntext = "Argument 'sign_doy1reference' is deprecated and will be removed soon. Use 'eofs_postprocessing_type' and 'eofs_postprocessing_params' instead."
        warnings.warn(warntext, DeprecationWarning, stacklevel=2)
        print(warntext)
    if interpolate_eofs is not None:
        deprecated_pp_interface_used = True
        warntext = "Argument 'interpolate_eofs' is deprecated and will be removed soon. Use 'eofs_postprocessing_type' and 'eofs_postprocessing_params' instead."
        warnings.warn(warntext, DeprecationWarning, stacklevel=2)
        print(warntext)
    if interpolation_start_doy is not None:
        deprecated_pp_interface_used = True
        warntext = "Argument 'interpolation_start_doy' is deprecated and will be removed soon. Use 'eofs_postprocessing_type' and 'eofs_postprocessing_params' instead."
        warnings.warn(warntext, DeprecationWarning, stacklevel=2)
        print(warntext)
    if interpolation_end_doy is not None:
        deprecated_pp_interface_used = True
        warntext = "Argument 'interpolation_end_doy' is deprecated and will be removed soon. Use 'eofs_postprocessing_type' and 'eofs_postprocessing_params' instead."
        warnings.warn(warntext, DeprecationWarning, stacklevel=2)
        print(warntext)
    if strict_leap_year_treatment is not None:
        warntext = "Argument 'strict_leap_year_treatment' is deprecated and will be removed soon. Use 'leap_year_treatment' instead."
        warnings.warn(warntext, DeprecationWarning, stacklevel=2)
        print(warntext)
        if strict_leap_year_treatment is True:
            leap_year_treatment = "strict"
        else:
            leap_year_treatment = "original"

    if deprecated_pp_interface_used:
        eofs_postprocessing_type = "kiladis2014"
        if sign_doy1reference is None:
            temp_sign_doy1reference = True
        else:
            temp_sign_doy1reference = sign_doy1reference
        if interpolate_eofs is None:
            temp_interpolate_eofs = False
        else:
            temp_interpolate_eofs = interpolate_eofs
        if interpolation_start_doy is None:
            temp_interpolation_start_doy = 293
        else:
            temp_interpolation_start_doy = interpolation_start_doy
        if interpolation_end_doy is None:
            temp_interpolation_end_doy = 316
        else:
            temp_interpolation_end_doy = interpolation_end_doy

        eofs_postprocessing_params = {"sign_doy1reference": temp_sign_doy1reference,
                                      "interpolate_eofs": temp_interpolate_eofs,
                                      "interpolation_start_doy": temp_interpolation_start_doy,
                                      "interpolation_end_doy": temp_interpolation_end_doy}

    # ###### end of backward compatibility section.


    preprocessed_olr = preprocess_olr(olrdata)
    raw_eofs = calc_eofs_from_preprocessed_olr(preprocessed_olr, implementation=implementation,
                                               leap_year_treatment=leap_year_treatment)
    result = initiate_eof_post_processing(raw_eofs, eofs_postprocessing_type, eofs_postprocessing_params)
    return result


def initiate_eof_post_processing(raw_eofs: eof.EOFDataForAllDOYs,
                                 eofs_postprocessing_type: str = "kiladis2014",
                                 eofs_postprocessing_params: dict = None) -> eof.EOFDataForAllDOYs:
    """
    Helper function that starts the selected post-processing routine and applies default post-processing parameters
    if none are given.

    :param raw_eofs: The EOFs, which have not been post-processed.
    :param eofs_postprocessing_type: see :py:func:`calc_eofs_from_olr`.
    :param eofs_postprocessing_params: see :py:func:`calc_eofs_from_olr`.

    :return: The post-processed EOFs.
    """
    if eofs_postprocessing_type is None:
        result = raw_eofs
    elif eofs_postprocessing_type == "kiladis2014":
        if eofs_postprocessing_params is None:
            eofs_postprocessing_params = {"sign_doy1reference": True,
                                          "interpolate_eofs": True,
                                          "interpolation_start_doy": 293,
                                          "interpolation_end_doy": 316}
        result = pp_kil2014.post_process_eofs_original_kiladis_approach(raw_eofs, **eofs_postprocessing_params)
    elif eofs_postprocessing_type == "eof_rotation":
        if eofs_postprocessing_params is None:
            eofs_postprocessing_params = {"sign_doy1reference": True}
        result = pp_rotation.post_process_eofs_rotation(raw_eofs, **eofs_postprocessing_params)

    else:
        raise ValueError("EOF post-processing type unknown.")
    return result

def preprocess_olr(olrdata: olr.OLRData) -> olr.OLRData:
    """
    Performs the preprocessing of an OLR dataset to make it suitable for the EOF analysis.

    This is a major step of the OMI algorithm and includes the Wheeler-Kiladis-Filtering.

    Note that it is recommended to use the function :py:func:`calc_eofs_from_olr` to cover the complete algorithm.

    :param olrdata: The original OLR dataset to be preprocessed. Note that OLR values are assumed to be given in
        positive values.

    :return: The filtered OLR dataset.
    """
    if np.mean(olrdata.olr) < 0:
        warnings.warn("OLR data apparently given in negative numbers. Here it is assumed that OLR is positive.")
    olrdata_filtered = wkfilter.filter_olr_for_mjo_eof_calculation(olrdata)
    return olrdata_filtered


def calc_eofs_from_preprocessed_olr(olrdata: olr.OLRData, implementation: str = "internal",
                                    leap_year_treatment: str = "original") -> eof.EOFDataForAllDOYs:
    """
    Calculates a series of EOF pairs: one pair for each DOY.

    This is based on already preprocessed OLR. Note that it is recommended to use the function
    :py:func:`calc_eofs_from_olr` to cover the complete algorithm.

    :param olrdata: the preprocessed OLR data, from which the EOFs are calculated.
    :param implementation: see :py:func:`calc_eofs_from_olr`.
    :param leap_year_treatment: see :py:func:`calc_eofs_from_olr`.
    :return: A pair of EOFs for each DOY. This series of EOFs has probably still to be postprocessed.
    """
    if implementation == "eofs_package" and not eofs_package_available:
        raise ValueError("Selected calculation with external eofs package, but package not available. Use "
                         "internal implementation or install eofs package")
    no_leap_years = False
    if leap_year_treatment == "no_leap_years":
        no_leap_years = True
    doys = tools.doy_list(no_leap_years)
    eofs = []
    for doy in doys:
        print("Calculating EOFs for DOY %i" % doy)
        if (implementation == "eofs_package"):
            singleeof = calc_eofs_for_doy_using_eofs_package(olrdata, doy, leap_year_treatment=leap_year_treatment)
        else:
            singleeof = calc_eofs_for_doy(olrdata, doy, leap_year_treatment=leap_year_treatment)
        eofs.append(singleeof)
    return eof.EOFDataForAllDOYs(eofs, no_leap_years)


def calc_eofs_for_doy(olrdata: olr.OLRData, doy: int, leap_year_treatment: str = "original") -> eof.EOFData:
    """
    Calculates a pair of EOFs for a particular DOY.

    An explicit internal implementation of the EOF approach is used.

    Note that it is recommended for the end user to call the function py:func:`calc_eofs_from_olr` to cover the
    complete algorithm.

    :param olrdata: The filtered OLR data to calculate the EOFs from.
    :param doy: The DOY for which the EOFs are calculated.
    :param leap_year_treatment: see :py:func:`calc_eofs_from_olr`.

    :return: An object containing the pair of EOFs together with diagnostic values.

    .. seealso:: :py:func:`calc_eofs_for_doy_using_eofs_package`

    """
    nlat = olrdata.lat.size
    nlong = olrdata.long.size
    olr_maps_for_doy = olrdata.extract_olr_matrix_for_doy_range(doy, window_length=60,
                                                                leap_year_treatment=leap_year_treatment)
    N = olr_maps_for_doy.shape[0]
    M = nlat * nlong
    F = np.reshape(olr_maps_for_doy, [N, M]).T  # vector: only one dimension. Length given by original longitude and latitude bins
    R = np.matmul(F, F.T) / N  # in some references, it is divided by (N-1), however, we follow Kutzbach (1967), in which it is only divided by N. In any case, the result should not differ much.
    if not np.allclose(R, R.T):
        warnings.warn("Covariance matrix is not symmetric within defined tolerance")
    L, E = np.linalg.eig(R)

    if not np.allclose(np.imag(L), 0.):
        warnings.warn("Imaginary part of at least one Eigenvalue greater than expected. Neglecting it anyway")
    L = np.real(L)
    order = (np.flip(L.argsort(), axis=None))
    L = L[order]
    total_var = np.sum(L)
    explainedVariances = L / total_var  # See Kutzbach (1967), Eq 12

    E = E[:, order]
    if not np.allclose(np.imag(E[:, 0:2]), 0.):
        warnings.warn("Imaginary part of one of the first two Eigenvectors greater than expected. Neglecting it anyway")
    E = np.real(E)
    eof1_vec = np.squeeze(E[:, 0])
    eof2_vec = np.squeeze(E[:, 1])

    return eof.EOFData(olrdata.lat, olrdata.long, eof1_vec, eof2_vec,
                       eigenvalues=L, explained_variances=explainedVariances, no_observations=N)


def calc_eofs_for_doy_using_eofs_package(olrdata: olr.OLRData, doy: int,
                                         leap_year_treatment: str = "original") -> eof.EOFData:
    """
    Calculates a pair of EOFs for a particular DOY.

    The external package :py:mod:`eofs` is used for the core calculation.

    Note that it is recommended to use the function :py:func:`calc_eofs_from_olr` to cover the complete algorithm.

    :param olrdata: The filtered OLR data to calculate the EOFs from.
    :param doy: The DOY for which the EOFs are calculated.
    :param leap_year_treatment: see :py:func:`calc_eofs_from_olr`.

    :return: An object containing the pair of EOFs together with diagnostic values.

    .. seealso:: :py:func:`calc_eofs_for_doy`

    """
    if eofs_package_available:
        nlat = olrdata.lat.size
        nlong = olrdata.long.size
        olr_maps_for_doy = olrdata.extract_olr_matrix_for_doy_range(doy, window_length=60,
                                                                    leap_year_treatment=leap_year_treatment)

        ntime = olr_maps_for_doy.shape[0]
        N = ntime
        M = nlat * nlong
        F = np.reshape(olr_maps_for_doy,
                       [N, M]).T  # vector: only one dimension. Length given by original longitude and latitude bins
        solver = eofs_package.Eof(F.T)
        # Todo: Should we care about complex values here?
        eofs = solver.eofs(neofs=2)
        explainedVariances = solver.varianceFraction()
        L = solver.eigenvalues()

        if L.size < M:
            # This usually happens if the covariance matrix did not have full rank (e.g. N<M). Missing Eigenvalues
            # are 0 and can be simply padded here
            L = np.pad(L, (0, M - L.size), 'constant', constant_values=(0, 0))
            explainedVariances = np.pad(explainedVariances, (0, M - explainedVariances.size), 'constant',
                                        constant_values=(0, 0))
        return eof.EOFData(olrdata.lat, olrdata.long, np.squeeze(eofs[0, :]), np.squeeze(eofs[1, :]),
                           eigenvalues=L, explained_variances=explainedVariances, no_observations=N)
    else:
        raise ModuleNotFoundError("eofs")

# #################PC Calculation

def calculate_pcs_from_olr(olrdata: olr.OLRData,
                           eofdata: eof.EOFDataForAllDOYs,
                           period_start: np.datetime64,
                           period_end: np.datetime64,
                           use_quick_temporal_filter=False) -> pc.PCData:
    """
    This major function computes PCs according to the OMI algorithm based on given OLR data and previously calculated
    EOFs.

    :param olrdata: The OLR dataset. The spatial grid must fit to that of the EOFs.
    :param eofdata: The previously calculated DOY-dependent EOFs.
    :param period_start: the beginning of the period, for which the PCs should be calculated.
    :param period_end: the ending of the period, for which the PCs should be calculated.
    :param use_quick_temporal_filter: There are two implementations of the temporal filtering
        available (the results of both methods are quite similar):

        * ``False``: the original Wheeler-Kiladis-Filter, which is closer to the original implementation while being
          slower (because it is based on a 2-dim FFT).
        * ``True``: 1-dim FFT Filter, which results in a quicker computation.

    :return: The PC time series. Normalized by the full PC time series
    """
    resticted_olr_data = olr.restrict_time_coverage(olrdata, period_start, period_end)
    resampled_olr_data = olr.interpolate_spatial_grid(resticted_olr_data, eofdata.lat, eofdata.long)
    if use_quick_temporal_filter:
        filtered_olr_data = qfilter.filter_olr_for_mjo_pc_calculation_1d_spectral_smoothing(resampled_olr_data)
    else:
        filtered_olr_data = wkfilter.filter_olr_for_mjo_pc_calculation(resampled_olr_data)
    raw_pcs = regress_3dim_data_onto_eofs(filtered_olr_data, eofdata)
    normalization_factor = 1 / np.std(raw_pcs.pc1)
    pc1 = np.multiply(raw_pcs.pc1, normalization_factor)
    pc2 = np.multiply(raw_pcs.pc2, normalization_factor)
    return pc.PCData(raw_pcs.time, pc1, pc2)


def calculate_pcs_from_olr_original_conditions(olrdata: olr.OLRData,
                                               original_eof_dirname: Path,
                                               use_quick_temporal_filter=False) -> pc.PCData:
    """
    Convenience function that calculates the OMI PCs for the original period using the original dataset (which has,
    however, to be provided by the user).

    :param olrdata: The original OLR data, which can be downloaded from
        https://www.psl.noaa.gov/thredds/catalog/Datasets/interp_OLR/catalog.html?dataset=Datasets/interp_OLR/olr.day.mean.nc. It can be read with the
        function :py:func:`mjoindices.olr_handling.load_noaa_interpolated_olr_netcdf4`.
    :param original_eof_dirname: Path to the original EOFs, which can be downloaded
        from ftp://ftp.cdc.noaa.gov/Datasets.other/MJO/eof1/ and ftp://ftp.cdc.noaa.gov/Datasets.other/MJO/eof2/ .
        The contents of both remote directories should again be placed into sub directories *eof1* and *eof2*
    :param use_quick_temporal_filter: See :py:func:`calculate_pcs_from_olr`

    :return: The PCs, which should be similar to the original ones.
    """
    period_start = np.datetime64("1979-01-01")
    period_end = np.datetime64("2018-08-28")
    eofs = eof.load_all_original_eofs_from_directory(original_eof_dirname)
    return calculate_pcs_from_olr(olrdata,
                                  eofs,
                                  period_start,
                                  period_end,
                                  use_quick_temporal_filter)


def regress_3dim_data_onto_eofs(data: object, eofdata: eof.EOFDataForAllDOYs) -> pc.PCData:
    """
    Finds time-dependent coefficients w.r.t the DOY-dependent EOF basis for time-dependent spatially resolved data.

    I.e. it finds the PCs for temporally resolved OLR data. But the function can also be used for other datasets,
    as long as those datasets have the same structure as the the class :class:`mjoindices.olr_handling.OLRData`.

    :param data: The data used to compute the coefficients. Should be an object of class
        :class:`mjoindices.olr_handling.OLRData` or of similar structure.
    :param eofdata: The DOY-dependent pairs of EOFs, as computed by, e.g., :func:`calc_eofs_from_olr`

    :return: The time-dependent PCs as :class:`mjoindices.principal_components.PCData`
    """
    if not np.all(data.lat == eofdata.lat):
        raise ValueError("Latitude grid of EOFs and OLR is not equal.")
    if not np.all(data.long == eofdata.long):
        raise ValueError("Longitude grid of EOFs and OLR is not equal.")
    pc1 = np.empty(data.time.size)
    pc2 = np.empty(data.time.size)

    for idx, val in enumerate(data.time):
        day = val
        olr_singleday = data.get_olr_for_date(day)
        doy = tools.calc_day_of_year(day, eofdata.no_leap_years)
        (pc1_single, pc2_single) = regress_vector_onto_eofs(
            eofdata.eofdata_for_doy(doy).reshape_to_vector(olr_singleday),
            eofdata.eof1vector_for_doy(doy),
            eofdata.eof2vector_for_doy(doy))
        pc1[idx] = pc1_single
        pc2[idx] = pc2_single
    return pc.PCData(data.time, pc1, pc2)


def regress_vector_onto_eofs(vector: np.ndarray, eof1: np.ndarray, eof2: np.ndarray) -> Tuple[float, float]:
    """
    Helper method that finds the coefficients of the given vector with respect to the given basis of 2 EOFs.

    The computed coefficients are the PCs in the terminology of the EOF analysis.

    :param vector: The vector for which the coefficients in the EOF basis should be found.
    :param eof1: EOF basis vector 1.
    :param eof2: EOF basis vector 2.

    :return: The two PCs.
    """
    eof_mat = np.array([eof1, eof2]).T

    # Alternative implementation 1:
    x = np.linalg.lstsq(eof_mat, vector, rcond=-1)
    pc1, pc2 = x[0]
    return pc1, pc2

    # Alternative implementation 2:
    # pseudo_inverse = np.linalg.pinv(eof_mat)
    # pcs = np.matmul(pseudo_inverse, vector)
    # return pcs[0], pcs[1]




# -*- coding: utf-8 -*-

# Copyright (C) 2019 Christoph G. Hoffmann. All rights reserved.

# This file is part of mjoindices

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Contact: christoph.hoffmann@uni-greifswald.de

"""
This module provides basic helper routines for the OMI calculation
"""

import datetime as dt
import typing
import numpy as np
import pandas as pd

def calc_day_of_year(date: typing.Union[np.datetime64, np.ndarray], no_leap_years: bool = False) -> typing.Union[int, np.ndarray]:
    """
    Calculates the days of the year (DOYs) for an individual date or an array of dates.

    :param date: The date (or the dates), given as (NumPy array of) :class:`numpy.datetime64` value(s).
    :param no_leap_years: ``True`` if every year has 365 days, ``False`` if dataset contains leap years.

    :return: the DOY (or the DOYs) as (NumPy array of) int value(s).
    """
    day_per_mon = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]

    if np.isscalar(date):
        if date.dtype == "<M8[ns]":
            # work around a bug, which prevents datetime64[ns] from being converted correctly to dt.datetime.
            date = date.astype("datetime64[us]")
   
        # expand datetime into individual time parts 
        temp = date.astype(dt.datetime)
        time_fragments = temp.timetuple()

        if no_leap_years:
            # check that day does not exceed number of days in a month
            if time_fragments.tm_mday > day_per_mon[time_fragments.tm_mon-1]:
                raise ValueError('Invalid date. Likely due to mismatch between input date and no_leap_years parameter')

            # sums days of previous months to get DOY
            result = sum(day_per_mon[:time_fragments.tm_mon-1]) + time_fragments.tm_mday
        else:
            result = time_fragments[7]
    else:
        result = np.empty(date.size)
        for i, d in enumerate(date):
            result[i] = calc_day_of_year(d, no_leap_years=no_leap_years)
    return result


def find_doy_ranges_in_dates(dates: np.ndarray, center_doy: int, window_length: int, leap_year_treatment: str = "original") -> typing.Tuple:
    """
    Finds the indices in a given array of dates that fit into a particular window of DOYs.

    This task sounds trivial, but is a little complicated by the appearance of leap years.

    :param dates: The array of dates as :class:`numpy.datetime64` values.
    :param center_doy: The center of the wanted window.
    :param window_length: the length of the window to both sides  of the center in days.
        The window spans 2*window_length+1 days in total (for exceptions see below).

    :param leap_year_treatment: Choose one of the following values:

        * ``"original"`` will be as close to the original version of :ref:`refKiladis2014` as possible. It will use a
          pragmatic implementation in which the start and end of the DOY window is directly computed as distance in
          units of DOYs.
        * ``"strict"`` (not recommended) will treat leap years somewhat more strictly, which might, however,
          cause the results to deviate from the original.  It will transfer the DOYs to actual calendar dates and
          will calculate the start and end of the window also as calendar dates using built-in numpy datetime functions.
        * ``"no_leap_years"`` will act as if there are no leap years in the dataset (365 days consistently),
          which might be useful for modeled data. It will follow the basic methodology of
          ``"original"``, but with ``ndoys = 365`` instead of potentially ``ndoys = 366``.

        In the context of the EOF calculation, the setting has major implications
        only for the EOFs calculated for DOY 366 and causes only minor differences for the other DOYs. The results for
        the setting ``"original"`` are closer to the original values, and approximately the same total number of calendar days
        covered by the window is found for each center DOY including DOY 366. However, the length of the window is not
        guaranteed to be ``2 * window_length + 1``, but can also be ``2 * window_length + 2`` if the window crosses the ending of a leap
        year. The setting ``"strict"`` is somewhat more stringently implemented. The window length is always ``2 * window_length + 1``,
        however, the number of calendar days covered by a window is reduced by approximately a factor of 4 for
        ``center_doy=366``, since a window is only found during leap years. This might cause the EOF to differ quite
        a lot from those of DOY 365 and DOY 1, which is not wanted. Because of that the recommended setting
        is the default value ``"original"``, except for modeled data that contains no leap years. In this case the setting
        ``"no_leap_years"`` is recommended.

    :return: Tuple with, first, the array of indices and, second, the resulting DOYs for comparison.

    """

    no_leap_years = False
    if leap_year_treatment == "no_leap_years":
        no_leap_years = True

    strict_leap_year_treatment = False
    if leap_year_treatment == "strict":
        strict_leap_year_treatment = True

    doys = calc_day_of_year(dates, no_leap_years=no_leap_years)

    if strict_leap_year_treatment:
        center_inds = np.nonzero(doys == center_doy)

        # switch from DOYs to real dates to use built-in leap year functionality
        startdates = dates[center_inds] - np.timedelta64(window_length, 'D')
        too_early_inds = np.nonzero(startdates < dates[0])
        startdates[too_early_inds] = dates[0]

        enddates = dates[center_inds] + np.timedelta64(window_length, 'D')
        too_late_inds = np.nonzero(enddates > dates[-1])
        enddates[too_late_inds] = dates[-1]

        resulting_idxlist = np.array([], dtype="int")
        for ind, startdate in enumerate(startdates):
            one_window_indices = np.nonzero((dates >= startdates[ind]) & (dates <= enddates[ind]))[0]
            resulting_idxlist = np.concatenate((resulting_idxlist, one_window_indices))
    else:
        lower_limit = center_doy - window_length
        if lower_limit < 1:
            lower_limit = lower_limit + 365
        upper_limit = center_doy + window_length
        if upper_limit > 365:
            upper_limit = upper_limit - 365

        if lower_limit <= upper_limit:
            inds_consider = ((doys >= lower_limit) & (doys <= upper_limit))
        else:
            inds_consider = ((doys >= lower_limit) | (doys <= upper_limit))
        resulting_idxlist = np.nonzero(inds_consider)[0]

    return np.asarray(resulting_idxlist), doys[resulting_idxlist]

def doy_list(no_leap_years: bool = False) -> np.array:

    """
    Returns an array of all DOYs in a year, hence simply the numbers from 1 to 365 or 366 (depending on the
    consideration of leap years). Useful, e.g., as axis for plotting.

    :param no_leap_years: ``True`` if every year has 365 days, ``False`` if dataset contains leap years.
    :return: The DOY array.
    """
    if no_leap_years:
        return np.arange(1, 366, 1) 
    else:
        return np.arange(1, 367, 1)







# -*- coding: utf-8 -*-

# Copyright (C) 2019 Christoph G. Hoffmann. All rights reserved.

# This file is part of mjoindices

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Contact: christoph.hoffmann@uni-greifswald.de

""" This module provides basic functionality to handle EOF data, which is a basic output of the OMI calculation. """

import copy
import typing
from pathlib import Path
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import matplotlib.cm
import warnings

from mjoindices.tools import doy_list


class EOFData:
    """
    This class serves as a container for the EOF data of one pair of EOFs.

    :param lat: The latitude grid of the EOF data.
    :param long: The longitude grid of the EOF data.
    :param eof1: Values of the first EOF. Can either be a 1-dim vector with
        ``lat.size * long.size`` elements (Start with all values of the first latitude, then all values of the second
        latitude, etc.) or a 2-dim map with the first index representing the latitude axis and the second index
        representing longitude axis.
    :param eof2: Values of the second EOF. Structure similar to ``eof1``
    :param explained_variances: Fraction of data variance that is explained by each EOF (for all EOFs and not only the
        first two EOFs). Can  be set to ``None``.
    :param eigenvalues: Eigenvalue corresponding to each EOF. Can  be set to ``None``.
    :param no_observations: The number of observations that went into the EOF calculation. Can  be set to ``None``.

    Note that the explained variances are not independent of the eigenvalues. However, this class is meant only to
    store the data. Hence, we store redundant data here intentionally to be able to have all computation performed together
    in another location.
    """

    def __init__(self, lat: np.ndarray, long: np.ndarray, eof1: np.ndarray, eof2: np.ndarray,
                 explained_variances: np.ndarray = None, eigenvalues: np.ndarray = None,
                 no_observations: int = None) -> None:
        """
        Initialization with all necessary variables.
        """
        if not eof1.shape == eof2.shape:
            raise ValueError("EOF1 and EOF2 must have the same shape")
        expected_n = lat.size * long.size

        if eof1.size != expected_n or eof2.size != expected_n:
            raise ValueError("Number of elements of EOF1 and EOF2 must be identical to lat.size*long.size")

        self._lat = lat.copy()
        self._long = long.copy()
        if eof1.ndim == 1 and eof2.ndim == 1:
            self._eof1 = eof1.copy()
            self._eof2 = eof2.copy()
        elif eof1.ndim == 2 and eof2.ndim == 2:
            if not (eof1.shape[0] == lat.size and eof1.shape[1] == long.size and eof2.shape[0] == lat.size and
                    eof2.shape[1]):
                raise ValueError("Length of first axis of EOS 1 and 2 must correspond to latitude axis, length of "
                                 "second axis to the longitude axis")
            self._eof1 = self.reshape_to_vector(eof1.copy())
            self._eof2 = self.reshape_to_vector(eof2.copy())
        else:
            raise ValueError("EOF1 and EOF2 must have a dimension of 1 or 2.")

        if eigenvalues is not None:
            if eigenvalues.size != self.eof1vector.size:
                raise ValueError("Eigenvalues (if not None) must have same length as the second axis of the EOFs")
            self._eigenvalues = eigenvalues.copy()
        else:
            self._eigenvalues = None
        if explained_variances is not None:
            if explained_variances.size != self.eof1vector.size:
                raise ValueError("Explained variances (if not None) must have same length as the second axis of "
                                 "the EOFs")
            self._explained_variances = explained_variances.copy()
        else:
            self._explained_variances = None
        self._no_observations = no_observations

    def __eq__(self, other: "EOFData") -> bool:
        """
        Override the default equals behavior
        """
        return (np.all(self.lat == other.lat)
                and np.all(self.long == other.long)
                and np.all(self.eof1vector == other.eof1vector)
                and np.all(self.eof2vector == other.eof2vector)
                and np.all(self._explained_variances == other.explained_variances)
                and np.all(self._eigenvalues == other.eigenvalues)
                and self._no_observations == other.no_observations)

    def close(self, other: "EOFData") -> bool:
        """
        Checks equality of two :class:`EOFData` objects, but allows for numerical tolerances.

        :param other: The object to compare with.

        :return: Equality of all members considering the default tolerances of :func:`numpy.allclose`
        """
        return (np.allclose(self.lat, other.lat)
                and np.allclose(self.long, other.long)
                and np.allclose(self.eof1vector, other.eof1vector)
                and np.allclose(self.eof2vector, other.eof2vector)
                and np.allclose(self._explained_variances, other.explained_variances)
                and np.allclose(self._eigenvalues, other.eigenvalues)
                and self._no_observations == other.no_observations)

    @property
    def lat(self) -> np.ndarray:
        """
        The latitude grid of the EOFs.
        """
        return self._lat

    @property
    def long(self) -> np.ndarray:
        """
        The longitude grid of the EOFs.
        """
        return self._long

    @property
    def eof1vector(self) -> np.ndarray:
        """
        EOF1 as a vector (and not a matrix).
        """
        return self._eof1

    @property
    def eof2vector(self) -> np.ndarray:
        """
        EOF2 as a vector (and not a matrix).
        """
        return self._eof2

    @property
    def eof1map(self) -> np.ndarray:
        """
        EOF1 as a 2-dimensional map.
        """
        return self.reshape_to_map(self._eof1)

    @property
    def eof2map(self) -> np.ndarray:
        """
        EOF2 as a 2-dimensional map.
        """
        return self.reshape_to_map(self._eof2)

    @property
    def explained_variances(self) -> np.ndarray:
        """
        The explained variances of all EOFs (not only the first two ones).

        :return: Explained Variances. Might be ``None``.
        """
        return self._explained_variances

    @property
    def explained_variance_eof1(self) -> float:
        """
        The explained variance of EOF1 as fraction between 0 and 1.

        :return: The variance. Might be ``None``.
        """
        if self._explained_variances is not None:
            return self._explained_variances[0]
        else:
            return None

    @property
    def explained_variance_eof2(self) -> float:
        """
        The explained variance of EOF2 as fraction between 0 and 1.

        :return: The variance. Might be ``None``.
        """
        if self._explained_variances is not None:
            return self._explained_variances[1]
        else:
            return None

    @property
    def sum_of_explained_variances(self) -> float:
        """
        Returns the total variance explained by all EOFs.

        This should be close to 1 if the calculation was successful.

        :return: The total explained variance. Might be ``None``.
        """
        if self._explained_variances is not None:
            return np.sum(self._explained_variances)
        else:
            return None

    @property
    def eigenvalues(self) -> np.ndarray:
        """
        The eigenvalues of all EOFs (not only the first two ones).

        :return: The eigenvalues. Might be ``None``.
        """
        return self._eigenvalues

    @property
    def eigenvalue_eof1(self) -> float:
        """
        The eigenvalue of EOF1.

        :return: The eigenvalue. Might be ``None``.
        """
        if self.eigenvalues is not None:
            return self._eigenvalues[0]
        else:
            return None

    @property
    def eigenvalue_eof2(self) -> float:
        """
        The eigenvalue of EOF2.

        :return: The eigenvalue. Might be ``None``.
        """
        if self.eigenvalues is not None:
            return self._eigenvalues[1]
        else:
            return None

    @property
    def no_observations(self) -> int:
        """
        The number of observations that went into the calculation of the EOFs.
        """
        return self._no_observations

    def reshape_to_vector(self, map: np.ndarray) -> np.ndarray:
        """
        Reshapes the horizontally distributed data to fit into a vector.

        The vector elements will contain the values of all longitudes for the first latitude,
        and then the values of all longitudes for the second latitude, etc.

        :param map: The 2-dim data. The first dimension must correspond to the latitude grid, the second to the
            longitude grid.

        :return: The data as a vector.
        """
        if not map.ndim == 2:
            raise ValueError("eof_map must have 2 dimensions")
        if not (map.shape[0] == self.lat.size and map.shape[1] == self.long.size):
            raise ValueError(
                "Length of first dimension of eof_map must correspond to the latitude grid, length of second dimension to the longitude grid")
        return np.reshape(map, self.lat.size * self.long.size)

    def reshape_to_map(self, vector: np.ndarray) -> np.ndarray:
        """
        Reshapes data in a vector to fit into a matrix, which corresponds to the present latitude/longitude grid.

        :param vector: The vector with the data. Must have the length ``lat.size * long.size``.

        :return: The map. The first index corresponds to the latitude grid, the second to longitude grid.
        """
        if not vector.ndim == 1:
            raise ValueError("Vector must have only 1 dimension.")
        if not vector.size == self.lat.size * self.long.size:
            raise ValueError("Vector must have lat.size*long.size elements.")
        # The following transformation has been double-checked graphically by comparing resulting plot to
        # https://www.esrl.noaa.gov/psd/mjo/mjoindex/animation/
        return np.reshape(vector, [self.lat.size, self.long.size])

    def save_eofs_to_txt_file(self, filename: Path) -> None:
        """
        Saves both EOFs to a .txt file.

        Note that the file format is not exactly that of the original data files. Particularly, both EOFs are written
        into the same file instead of 2 separate files. Furthermore, the lat/long-grids are also explicitly saved.

        A suitable reader is provided by the function
        :py:func:`load_single_eofs_from_txt_file`, whereas a reader for the
        original files is provided by :py:func:`load_original_eofs_for_doy`

        :param filename: The full path- and filename.
        """
        lat_full = np.empty(self.eof1vector.size)
        long_full = np.empty(self.eof1vector.size)
        for i_vec in range(0, self.eof1vector.size):
            # find out lat/long corresponding to the vector position by evaluating the index transformation from map
            # to vector
            (i_lat, i_long) = np.unravel_index(i_vec, self.eof1map.shape)
            lat_full[i_vec] = self.lat[i_lat]
            long_full[i_vec] = self.long[i_long]
        df = pd.DataFrame(
            {"Lat": lat_full, "Long": long_full, "EOF1": self.eof1vector, "EOF2": self.eof2vector}).astype(float)
        df.to_csv(filename, index=False, float_format="%13.7f")


class EOFDataForAllDOYs:
    """
    This class serves as a container for a series of EOF pairs, which covers all DOYs and provides some overall
    statistical quantities.

    The basic EOF computation function :py:func:`mjoindices.omi.omi_calculator.calc_eofs_from_olr` will return an object
    of this class as a major result of this package.

    The individual EOF pairs are represented by :class:`EOFData` objects.

    Note that the user can choose to consider or to ignore leap years in the data. This is probably of interest for the work
    with modeled data, since models might ignore leap years in idealized setups. Here, this means that the number of
    considered DOYs (and therefore the number of EOF pairs and further variables) can be either 365 or 366.
    The code in this package is designed to treat both cases consistently, however, the users should have their own
    choice in mind when working with the results.

    :param eof_list: A list with one :class:`EOFData` object for each DOY.
    :param no_leap_years: ``True`` if every year has 365 days, ``False`` if dataset contains leap years.
    """

    def __init__(self, eof_list: typing.List[EOFData], no_leap_years: bool) -> None:
        if (no_leap_years and len(eof_list) != 365):
            raise ValueError("List of EOFs must contain 365 entries for no_leap_years=True")
        elif (not no_leap_years and len(eof_list) != 366):
            raise ValueError("List of EOFs must contain 366 entries for no_leap_years=False") 

        if no_leap_years: 
            reference_list_len = 365 # number of days in year
        else:
            reference_list_len = 366
        reference_lat = eof_list[0].lat
        reference_long = eof_list[0].long
        for i in range(0, reference_list_len):
            if not np.all(eof_list[i].lat == reference_lat):
                raise ValueError("All EOFs must have the same latitude grid. Problematic is DOY %i" % i)
            if not np.all(eof_list[i].long == reference_long):
                raise ValueError("All EOFs must have the same longitude grid. Problematic is DOY %i" % i)
        # deepcopy eofs so that they cannot be modified accidentally from outside after the consistency checks
        self._eof_list = copy.deepcopy(eof_list)
        self._no_leap_years = no_leap_years

    # FIXME: Could this property be used to modify the EOFData objects because they are mutual?
    @property
    def eof_list(self) -> typing.List[EOFData]:
        """
        EOF data for all DOYs as a list.

        Remember that DOY 1 corresponds to list entry 0.
        """
        return self._eof_list

    @property
    def no_leap_years(self) -> bool:
        """
        ``True`` if every year has 365 days, ``False`` if dataset contains leap years.
        """
        return self._no_leap_years

    @property
    def lat(self) -> np.ndarray:
        """
        The latitude grid common to the EOFs of all DOYs.

        """
        return self.eof_list[0].lat

    @property
    def long(self) -> np.ndarray:
        """
        The longitude grid common to the EOFs of all DOYs.
        """
        return self.eof_list[0].long

    @property
    def len_eof_list(self) -> int:
        """
        Length of the ``eof_list``, based on ``no_leap_years``, i.e. the number of considered DOYs.
        """
        if self._no_leap_years:
            return 365
        else:
            return 366

    def eofdata_for_doy(self, doy: int) -> EOFData:
        """
        Returns the :class:`EOFData` object for a particular DOY.

        :param doy: The DOY

        :return: The :class:`EOFData` object
        """
        return self.eof_list[doy - 1]

    def eof1vector_for_doy(self, doy: int) -> np.ndarray:
        """
        Shortcut to the EOF1 vector of a particular DOY.

        :param doy: The DOY.

        :return: The vector.
        """
        return self.eof_list[doy - 1].eof1vector

    def eof2vector_for_doy(self, doy: int) -> np.ndarray:
        """
        Shortcut to the EOF2 vector of a particular DOY.

        :param doy: The DOY.

        :return: The vector.
        """
        return self.eof_list[doy - 1].eof2vector

    def explained_variance1_for_all_doys(self):
        """
        Returns a vector containing the explained variance of EOF1 for each DOY.

        :return: The variance vector.

        """
        doys = doy_list(self._no_leap_years)
        result = []
        for doy in doys:
            result.append(self.eofdata_for_doy(doy).explained_variance_eof1)
        return np.array(result)

    def explained_variance2_for_all_doys(self):
        """
        Returns a vector containing the explained variance of EOF2 for each DOY.

        :return: The variance vector.
        """
        doys = doy_list(self._no_leap_years)
        result = []
        for doy in doys:
            result.append(self.eofdata_for_doy(doy).explained_variance_eof2)
        return np.array(result)

    def total_explained_variance_for_all_doys(self):
        """
        Returns a vector containing -for each DOY- the sum of the explained variance over all EOFs.

        :return: The variance vector. Should by close to 1 for each DOY if computation was successful.
        """
        doys = doy_list(self._no_leap_years)
        result = []
        for doy in doys:
            result.append(self.eofdata_for_doy(doy).sum_of_explained_variances)
        return np.array(result)

    def no_observations_for_all_doys(self):
        """
        Returns a vector containing -for each DOY- the number of observations that went into the
        computation of the EOFs.

        :return: The number of observations vector.
        """
        doys = doy_list(self._no_leap_years)
        result = []
        for doy in doys:
            result.append(self.eofdata_for_doy(doy).no_observations)
        return np.array(result)

    def eigenvalue1_for_all_doys(self):
        """
        Returns a vector containing the eigenvalues of EOF1 for each DOY.

        :return: The eigenvalue vector.
        """
        doys = doy_list(self._no_leap_years)
        result = []
        for doy in doys:
            result.append(self.eofdata_for_doy(doy).eigenvalue_eof1)
        return np.array(result)

    def eigenvalue2_for_all_doys(self):
        """
        Returns a vector containing the eigenvalues of EOF2 for each DOY.

        :return: The eigenvalue vector.
        """
        doys = doy_list(self._no_leap_years)
        result = []
        for doy in doys:
            result.append(self.eofdata_for_doy(doy).eigenvalue_eof2)
        return np.array(result)

    def save_all_eofs_to_dir(self, dirname: Path, create_dir=True) -> None:
        """
        Saves the EOF1 and EOF2 data for each of the DOYs in the given directory.

        For each DOY, one text file will be created, which contains both EOFs.
        Note that the text files do not contain the eigenvalues and explained variance values. To save also those
        values, use the function :py:func:`save_all_eofs_to_npzfile`.

        :param dirname: The directory, where the files will be saved into.
        :param create_dir: If ``True``, the directory (and parent directories) will be created, if not existing.
        """
        if not dirname.exists() and create_dir:
            dirname.mkdir(parents=True, exist_ok=False)
        for doy in doy_list(self.no_leap_years):
            filename = dirname / Path("eof%s.txt" % format(doy, '03'))
            self.eofdata_for_doy(doy).save_eofs_to_txt_file(filename)

    def save_all_eofs_to_npzfile(self, filename: Path) -> None:
        """
        Saves the complete EOF data to a numpy file.

        :param filename: The filename.
        """
        doys = doy_list(self._no_leap_years)
        eof1 = np.empty((doys.size, self.lat.size * self.long.size))
        eof2 = np.empty((doys.size, self.lat.size * self.long.size))
        eigenvalues = np.empty((doys.size, self.lat.size * self.long.size))
        explained_variances = np.empty((doys.size, self.lat.size * self.long.size))
        no_observations = np.empty(doys.size)
        for i in range(0, doys.size):
            eof = self.eof_list[i]
            eof1[i, :] = eof.eof1vector
            eof2[i, :] = eof.eof2vector
            eigenvalues[i, :] = eof.eigenvalues
            explained_variances[i, :] = eof.explained_variances
            no_observations[i] = eof.no_observations
        np.savez(filename,
                 eof1=eof1,
                 eof2=eof2,
                 explained_variances=explained_variances,
                 eigenvalues=eigenvalues,
                 no_observations=no_observations,
                 lat=self.lat,
                 long=self.long)


def load_single_eofs_from_txt_file(filename: Path) -> EOFData:
    """
    Loads a pair of EOFs, which was previously saved with this package (function :py:func:`EOFData.save_eofs_to_txt_file`).

    :param filename: Path to the  EOF file.

    :return: The pair of EOFs.
    """
    df = pd.read_csv(filename, sep=',', header=0)
    full_lat = df.Lat.values
    full_long = df.Long.values
    eof1 = df.EOF1.values
    eof2 = df.EOF2.values

    # retrieve unique lat/long grids
    # FIXME: Does probably not work for a file with only one lat
    lat, repetition_idx, rep_counts = np.unique(full_lat, return_index=True, return_counts=True)
    long = full_long[repetition_idx[0]:repetition_idx[1]]

    # Apply some heuristic consistency checks
    if not np.unique(rep_counts).size == 1:
        # All latitudes must have the same number of longitudes
        raise ValueError("Lat/Long grid in input file seems to be corrupted 1")
    if not np.unique(repetition_idx[0:-1] - repetition_idx[1:]).size == 1:
        # All beginnings of a new lat have to be equally spaced
        raise ValueError("Lat/Long grid in input file seems to be corrupted 2")
    if not lat.size * long.size == eof1.size:
        raise ValueError("Lat/Long grid in input file seems to be corrupted 3")
    if not np.all(np.tile(long, lat.size) == full_long):
        # The longitude grid has to be the same for all latitudes
        raise ValueError("Lat/Long grid in input file seems to be corrupted 4")

    return EOFData(lat, long, eof1, eof2)


def load_original_eofs_for_doy(dirname: Path, doy: int) -> EOFData:
    """
    Loads the EOF values for the first 2 EOFs from the original file format.

    Note that the EOFs are represented as pure vectors in the original treatment, so that a connection to the
    individual locations on a world map is not obvious without any further knowledge. The corresponding grid is 
    hardcoded in here.

    The original EOFs are found here: ftp://ftp.cdc.noaa.gov/Datasets.other/MJO/eof1/ and
    ftp://ftp.cdc.noaa.gov/Datasets.other/MJO/eof2/

    :param dirname: Path to the directory, in which the EOFs for all DOYs are stored.
        This path should contain the sub directories "eof1" and "eof2", in which the 365 or 366 files each are located:
        One file per DOY.
    :param doy: DOY for which the 2 EOFs are loaded (number between 1 and 366).

    :return: The pair of EOFs.
    """
    orig_lat = np.arange(-20., 20.1, 2.5)
    orig_long = np.arange(0., 359.9, 2.5)
    eof1filename = dirname / "eof1" / ("eof" + str(doy).zfill(3) + ".txt")
    eof1 = np.genfromtxt(eof1filename)
    eof2filename = dirname / "eof2" / ("eof" + str(doy).zfill(3) + ".txt")
    eof2 = np.genfromtxt(eof2filename)
    return EOFData(orig_lat, orig_long, eof1, eof2)


def load_all_eofs_from_directory(dirname: Path) -> EOFDataForAllDOYs:
    """
    Loads the EOF functions (created with the function :py:func:`EOFDataForAllDOYs.save_all_eofs_to_dir`)
    for all DOYs from the given directory

    :param dirname: The directory in which the files are stored.

    :return: The EOFs for all DOYs.
    """
    eofs = []
    for doy in doy_list(no_leap_years=False):
        filename = dirname / Path("eof%s.txt" % format(doy, '03'))
        if doy < 366:
            eof = load_single_eofs_from_txt_file(filename)
            eofs.append(eof) 
        else:
            # try to load DOY 366 from directory, if it exists. 
            try:
                eof = load_single_eofs_from_txt_file(filename)
                eofs.append(eof)
                no_leap_years = False
            except:
                no_leap_years = True
                warnings.warn('No EOFs from DOY 366 in directory. Assuming no leap years in dataset.')            
    return EOFDataForAllDOYs(eofs, no_leap_years) 


def load_all_original_eofs_from_directory(dirname: Path) -> EOFDataForAllDOYs:
    """
    Loads the EOF functions for all DOYs from the original file format.

    The original EOFs are found here: ftp://ftp.cdc.noaa.gov/Datasets.other/MJO/eof1/ and
    ftp://ftp.cdc.noaa.gov/Datasets.other/MJO/eof2/

    Note that the EOFs are represented as pure vectors in the original treatment, so that a connection to the
    individual locations on a world map is not obvious without any further knowledge. The corresponding grid is
    hardcoded in here.

    :param dirname: Path to the directory, in which the EOFs for all DOYs are stored.
        This path should contain the sub directories "eof1" and "eof2", in which the 366 files each are located:
        One file per day of the year.

    :return: The original EOFs for all DOYs.
    """
    eofs = []
    for doy in doy_list(no_leap_years=False): # the original files contain leap years
        eof = load_original_eofs_for_doy(dirname, doy)
        eofs.append(eof)
    return EOFDataForAllDOYs(eofs, no_leap_years=False)


def restore_all_eofs_from_npzfile(filename: Path) -> EOFDataForAllDOYs:
    """
    Loads all EOF data from a numpy file, which was written with :py:func:`EOFDataForAllDOYs.save_all_eofs_to_npzfile`.

    :param filename: The filename.

    :return: The EOFs for all DOYs.
    """
    with np.load(filename) as data:
        eof1 = data["eof1"]
        eof2 = data["eof2"]
        lat = data["lat"]
        long = data["long"]
        eigenvalues = data["eigenvalues"]
        explained_variances = data["explained_variances"]
        no_observations = data["no_observations"]
    eofs = []
    if eof1.shape[0] == 365:
        no_leap_years = True # sets True if no leap years in dataset. 
    elif eof1.shape[0] == 366:
        no_leap_years = False
    else:
        raise ValueError('Dataset does not have EOFs from each day of year.')

    for i in range(0, doy_list(no_leap_years).size):
        eof = EOFData(lat, long, np.squeeze(eof1[i, :]), np.squeeze(eof2[i, :]),
                      eigenvalues=np.squeeze(eigenvalues[i, :]),
                      explained_variances=np.squeeze(explained_variances[i, :]),
                      no_observations=no_observations[i])
        eofs.append(eof)
    return EOFDataForAllDOYs(eofs, no_leap_years)


def plot_explained_variance_for_all_doys(eofs: EOFDataForAllDOYs, include_total_variance: bool = False,
                                         include_no_observations: bool = False) -> Figure:
    """
    Plots the explained variance values for EOF1 and EOF2 for all DOYs.

    Comparable to :ref:`refKiladis2014`, Fig. 1 (although the values there are too high by a factor of 2). See
    :ref:`refKiladis2020` for the correct original values.

    :param eofs: The EOF data to plot.

    :return: Handle to the figure.
    """
    doygrid = doy_list(eofs.no_leap_years)
    fig = plt.figure("plot_explained_variance_for_all_doys", clear=True, figsize=(6, 4), dpi=150)
    ax1 = fig.add_subplot(111)
    handles = []
    p1, = ax1.plot(doygrid, eofs.explained_variance1_for_all_doys(), color="blue", label="EOF1")
    handles.append(p1)
    p2, = ax1.plot(doygrid, eofs.explained_variance2_for_all_doys(), color="red", label="EOF2")
    handles.append(p2)
    if include_total_variance:
        p3, = ax1.plot(doygrid, eofs.total_explained_variance_for_all_doys(), color="green", label="Total")
        handles.append(p3)
    ax1.set_xlabel("DOY")
    ax1.set_ylabel("Fraction of explained variance")
    ax1.set_xlim((0, eofs.len_eof_list))

    if include_no_observations:
        ax2 = ax1.twinx()
        p4, = ax2.plot(doygrid, eofs.no_observations_for_all_doys(), color="black", label="Number of observations", linestyle="--")
        handles.append(p4)
        ax2.set_ylabel("Number of observations")
        ymin = np.min(eofs.no_observations_for_all_doys()) - np.min(eofs.no_observations_for_all_doys()) * 0.1
        ymax = np.max(eofs.no_observations_for_all_doys()) + np.max(eofs.no_observations_for_all_doys()) * 0.1
        ax2.set_ylim([ymin, ymax])
    plt.title("Explained variance")
    plt.legend(handles=tuple(handles))
    return fig


def plot_eigenvalues_for_all_doys(eofs: EOFDataForAllDOYs) -> Figure:
    """
    Plots the Eigenvalues for EOF1 and EOF2 for all DOYs.

    :param eofs: The EOF data to plot.

    :return: Handle to the figure.
    """
    doygrid = doy_list(eofs.no_leap_years)
    fig = plt.figure("plot_eigenvalues_for_all_doys", clear=True, figsize=(6, 4), dpi=150)
    p1, = plt.plot(doygrid, eofs.eigenvalue1_for_all_doys(), color="blue", label="EOF1")
    p2, = plt.plot(doygrid, eofs.eigenvalue2_for_all_doys(), color="red", label="EOF2")
    plt.xlabel("DOY")
    plt.ylabel("Eigenvalue")
    plt.title("Eigenvalues")
    plt.legend(handles=(p1, p2))
    return fig


def plot_original_individual_eof_map(path, doy: int) -> Figure:
    """
    Plots a pair of original EOFs, which are loaded from a directory, in two maps.

    :param path: The directory with the EOF data (see :py:func:`load_original_eofs_for_doy` for details).
    :param doy: The corresponding DOY. Only used to display it in the title.

    :return: Handle to the figure.
    """
    eofdata = load_original_eofs_for_doy(path, doy)
    return plot_individual_eof_map(eofdata, doy=doy)


def plot_individual_eof_map_from_file(filename, doy: int) -> Figure:
    """
    Plots a pair of EOFs, which are loaded from a file, in two maps.

    :param filename: The file with the EOF data.
    :param doy: The corresponding DOY. Only used to display it in the title.

    :return: Handle to the figure.
    """
    eofdata = load_single_eofs_from_txt_file(filename, doy=doy)
    return plot_individual_eof_map(eofdata)


def plot_individual_eof_map(eofdata: EOFData, doy: int = None) -> Figure:
    """
    Plots a pair of EOFs for a particular DOY in two maps.

    :param eofdata: The EOF data to plot.
    :param doy: The corresponding DOY. Only used to display it in the title.

    :return: Handle to the figure.
    """
    # TODO: Plot underlying map
    fig, axs = plt.subplots(2, 1, num="plotting.plot_eof_for_doy", clear=True,
                            figsize=(10, 5), dpi=150, sharex=True, sharey=True)
    plt.subplots_adjust(wspace=0.35, hspace=0.35)
    if doy is not None:
        fig.suptitle("EOF Recalculation for DOY %i" % doy)

    ax = axs[0]

    c = ax.contourf(eofdata.long, eofdata.lat, eofdata.eof1map, levels=np.arange(-0.1, 0.11, 0.01), cmap=matplotlib.cm.get_cmap("bwr"))
    fig.colorbar(c, ax=ax, label="OLR Anomaly [W/m²]")
    ax.set_title("EOF1")
    ax.set_ylabel("Latitude [°]")
    ax.set_xlabel("Longitude [°]")

    ax = axs[1]
    c = ax.contourf(eofdata.long, eofdata.lat, eofdata.eof2map, levels=np.arange(-0.1, 0.11, 0.01), cmap=matplotlib.cm.get_cmap("bwr"))
    fig.colorbar(c, ax=ax, label="OLR Anomaly [W/m²]")
    ax.set_title("EOF2")
    ax.set_ylabel("Latitude [°]")
    ax.set_xlabel("Longitude [°]")

    return fig


def plot_individual_explained_variance_all_eofs(eof: EOFData, doy: int = None, max_eof_number: int = None) -> Figure:
    """
    Plots the explained variances for each EOF function, but only for EOF data of one DOY.

    This is useful to confirm that the first 2 EOFs cover actually most of the variance.

    :param eof: The EOF data.
    :param doy: The corresponding DOY. Only used to display it in the title.
    :param max_eof_number: The limit of the x-axis.

    :return: Handle to the figure.
    """
    eof_number_grid = np.arange(0, eof.explained_variances.size, 1) + 1
    fig = plt.figure("plot_individual_explained_variance_all_eofs", clear=True, figsize=(6, 4), dpi=150)
    ax = fig.add_subplot(111)
    p1 = plt.plot(eof_number_grid, eof.explained_variances, color="blue", label="Explained variance")
    if max_eof_number is not None:
        plt.xlim(0.5, max_eof_number)
    plt.xlabel("EOF Number")
    plt.ylabel("Fraction of explained variance")
    plt.text(0.6, 0.9, "Sum over all EOFs: %1.2f" % eof.sum_of_explained_variances, transform=ax.transAxes)

    if doy is not None:
        plt.title("Explained variance for DOY %i" % doy)
    else:
        plt.title("Explained variance")
    return fig




# -*- coding: utf-8 -*-

# Copyright (C) 2019 Christoph G. Hoffmann. All rights reserved.

# This file is part of mjoindices

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Contact: christoph.hoffmann@uni-greifswald.de

"""
This module provides basic functionality to handle OLR data, which is the basic input for the OMI calculation.
"""

from pathlib import Path

import numpy as np
import scipy
import scipy.interpolate
from matplotlib.figure import Figure
import scipy.io
import netCDF4 as netcdf4
import matplotlib.pyplot as plt

import mjoindices.tools as tools


class OLRData:
    """
    This class serves as a container for spatially distributed and temporally resolved OLR data.

    A filled object of this class has to be provided by the user in order to start the OMI calculation.

    :param olr: The OLR data as a 3-dim array. The three dimensions correspond to time, latitude, and longitude, in this
        order.
    :param time: The temporal grid as 1-dim array of :class:`numpy.datetime64` dates.
    :param lat: The latitude grid as 1-dim array.
    :param long: The longitude grid as 1-dim array.
    """

    def __init__(self, olr: np.ndarray, time: np.ndarray, lat: np.ndarray, long: np.ndarray) -> None:
        """
        Initialization of basic variables.
        """
        if olr.shape[0] != time.size:
            raise ValueError('Length of time grid does not fit to first dimension of OLR data cube')
        if olr.shape[1] != lat.size:
            raise ValueError('Length of lat grid does not fit to second dimension of OLR data cube')
        if olr.shape[2] != long.size:
            raise ValueError('Length of long grid does not fit to third dimension of OLR data cube')
        self._olr = olr.copy()
        self._time = time.copy()
        self._lat = lat.copy()
        self._long = long.copy()

    @property
    def olr(self):
        """
        The OLR data as a 3-dim array. The three dimensions correspond to time, latitude, and longitude, in this
        order.
        """
        return self._olr

    @property
    def time(self):
        """
        The temporal grid as 1-dim array of :class:`numpy.datetime64` dates.
        """
        return self._time

    @property
    def lat(self):
        """
        The latitude grid as 1-dim array.
        """
        return self._lat

    @property
    def long(self):
        """
        The longitude grid as 1-dim array.
        """
        return self._long

    def __eq__(self, other: "OLRData") -> bool:
        """
        Override the default Equals behavior
        """
        return (np.all(self.lat == other.lat)
                and np.all(self.long == other.long)
                and np.all(self.time == other.time)
                and np.all(self.olr == other.olr))

    def close(self, other: "OLRData") -> bool:
        """
         Checks equality of two :py:class:`OLRData` objects, but allows for numerical tolerances.

        :param other: The object to compare with.

        :return: Equality of all members considering the default tolerances of :py:func:`numpy.allclose`
        """
        return (np.allclose(self.lat, other.lat)
                and np.allclose(self.long, other.long)
                and np.allclose(self.time.astype("float"), other.time.astype("float"))  # allclose does not work with datetime64
                and np.allclose(self.olr, other.olr))

    def get_olr_for_date(self, date: np.datetime64) -> np.ndarray:
        """
        Returns the spatially distributed OLR map for a particular date.

        :param date: The date, which has to be exactly matched by one of the dates in the OLR time grid.

        :return: The excerpt of the OLR data as a 2-dim array. The two dimensions correspond to
            latitude and longitude, in this order. Returns ``None`` if the date is not contained in the OLR time series.
        """
        cand = self.time == date
        if not np.all(cand == False):  # noqa: E712
            return np.squeeze(self.olr[cand, :, :])
        else:
            return None

    def extract_olr_matrix_for_doy_range(self, center_doy: int, window_length: int = 0, leap_year_treatment: str = "original") -> np.ndarray:
        """
        Extracts a range of OLR data from the DOYs around one center DOY (center_doy +/- windowlength).

        Keep in mind that the OLR time series might span several years. In this case the center DOY is found more than
        once and the respective window in considered for each year.
        Example: 3 full years of data, centerdoy = 20, and window_length = 4 results in 3*(2*4+1) = 27 entries in the
        time axis

        :param center_doy: The center DOY of the window.
        :param window_length: The window length of DOYs on both sides of the center DOY. Hence, if the window is fully
            covered by the data, one gets :math:``2*window_length + 1`` entries per year in the result.
        :param leap_year_treatment: see :py:func:``mjoindices.omi.omi_calculator.calc_eofs_from_olr``.
        :return: The excerpt of the OLR data as a 3-dim array. The three dimensions correspond to
            time, latitude, and longitude, in this order.

        """
        inds, doys = tools.find_doy_ranges_in_dates(self.time, center_doy, window_length=window_length,
                                                    leap_year_treatment=leap_year_treatment)
        return self.olr[inds, :, :]

    def save_to_npzfile(self, filename: Path) -> None:
        """
        Saves the data arrays contained in the OLRData object to a numpy file.

        :param filename: The full filename.
        """
        np.savez(filename, olr=self.olr, time=self.time, lat=self.lat, long=self.long)


def interpolate_spatial_grid_to_original(olr: OLRData) -> OLRData:
    """
    Convenience function that interpolates the OLR data in an :class:`OLRData` object spatially onto the spatial grid,
    which was used for the original OMI calculation by :ref:`refKiladis2014`.

    This original grid has the following properties:

    * Latitude: 2.5 deg-sampling in the tropics from -20 to 20 deg (20 S to 20 N).
    * Longitude: Whole globe with 2.5 deg-sampling.

    :param olr: The OLR data

    :return:  A new :class:`OLRData` object with the interpolated data.
    """
    # FIXME Combine with definition in empirical_or....py
    orig_lat = np.arange(-20., 20.1, 2.5)
    orig_long = np.arange(0., 359.9, 2.5)
    return interpolate_spatial_grid(olr, orig_lat, orig_long)


def interpolate_spatial_grid(olr: OLRData, target_lat: np.ndarray, target_long: np.ndarray) -> OLRData:
    """
    Interpolates the OLR data linearly onto the given grids.

    No extrapolation will be done. Instead, a :py:class:`ValueError` is raised if the data does not cover the target
    grid.

    Note that no sophisticated resampling is provided here. If some kind of averaging, etc., is needed, it should
    be performed by the user before injecting the data into the OMI calculation.

    :param olr: The OLR data to resample.
    :param target_lat: The new latitude grid.
    :param target_long: The new longitude grid.

    :return: A new :class:`OLRData` object containing the resampled OLR data.
    """
    no_days = olr.time.size
    olr_interpol = np.empty((no_days, target_lat.size, target_long.size))
    for idx in range(0, no_days):
        f = scipy.interpolate.interp2d(olr.long, olr.lat, np.squeeze(olr.olr[idx, :, :]), kind='linear', bounds_error=True)
        olr_interpol[idx, :, :] = f(target_long, target_lat)
    return OLRData(olr_interpol, olr.time, target_lat, target_long)


def restrict_time_coverage(olr: OLRData, start: np.datetime64, stop: np.datetime64) -> OLRData:
    """
    Cuts the OLR time series at the given dates (given dates are included).

    This is useful when the OLR data should be restricted for the EOF calculation, which is based on a subset.
    Of course, it can also be used to limit the PC calculation to a specific period.

    Note that a temporal resampling method is not provided here, since the possible resampling methods,
    which the users might want to apply, are too diverse. Hence, it is assumed that the temporal spacing is already
    correct (daily averages recommended) and only a restriction of the period is needed before calculation.

    :param olr: The OLR data to restrict.
    :param start: The beginning of the wanted period of OLR data (included).
    :param stop: The ending of the wanted period (included).

    :return: A new :class:`OLRData` object with restricted temporal coverage.

    :raises: :py:class:`ValueError` if no OLR Data is found for the specified period
    """
    window_inds = (olr.time >= start) & (olr.time <= stop)
    if np.all(window_inds == False):  # noqa: E712
        raise ValueError("No OLR data within specified period found. Data covers the period from %s to %s."
                         % (str(olr.time[0]), str(olr.time[-1])))
    else:
        return OLRData(olr.olr[window_inds, :, :], olr.time[window_inds], olr.lat, olr.long)


def remove_leap_years(olr: OLRData) -> OLRData:
    """
    Removes any leap days (any instances of February 29) and returns an :class:`OLRData` object
    with the remaining OLR data

    :param olr: The OLR data to restrict.

    :return: A new :class:`OLRData` object with no leap days
    """

    window_inds = [(i.astype(object).month != 2) | (i.astype(object).day != 29) for i in olr.time]

    return OLRData(olr.olr[window_inds, :, :], olr.time[window_inds], olr.lat, olr.long) 
    


def load_noaa_interpolated_olr(filename: Path, use_xarray: bool = False) -> OLRData:
    """
    Loads the standard OLR data product provided by NOAA in NetCDF3 format.
    This is mainly used to load the OLR files originally used for the OMI calculation some years ago.

    ATTENTION: Note that the file format was changed by NOAA from NetCDF3 to NetCDF4 sometime
    between the years 2019 and 2021. If you are using a recent download of the data and experience problems
    with this loader method, you should use :py:func:`load_noaa_interpolated_olr_netcdf4` instead.

    The original OLR data file is contained in the reference data package found at: https://doi.org/10.5281/zenodo.3746562

    The current dataset can be obtained from
    https://www.psl.noaa.gov/thredds/catalog/Datasets/interp_OLR/catalog.html?dataset=Datasets/interp_OLR/olr.day.mean.nc

    A description is found at
    https://psl.noaa.gov/data/gridded/data.olrcdr.interp.html

    :param filename: Full filename of a local copy of OLR data file.
    :param use_xarray: Option to use xarray instead of SciPy netcdf for faster loading of data and timestamps. Note that OMI values
        based on this parameter have not been validated. The respective unit tests currently fail when this option is activated,
        however, this is probably only caused by very small numerical differences, which are irrelevant for the actual use.
        Still make sure that you trust the values when activating the option. A further advantage of this option is that is works for
        NetCDF3 and NetCDF4 files, hence for the older and the newer NOAA OLR datafiles.

    :return: The OLR data.
    """
    # ToDo: Validate the complete OMI calculation chain with OLR data loaded using the xarray option (Quick tests showed
    # that the OMI values are probably essentially the same but numerical differeces lead to failiures of the integration tests).
    # If the values are ok, make the xarray option the default (which also means that Python 3.8 is at least needed) and remove the
    # special function for NetCDF 4 below.
    if use_xarray:
        import xarray as xr
        f = xr.open_dataset(filename)
        lat = f.lat.data.copy()
        lon = f.lon.data.copy()
        olr = f.olr.data.copy()
        time = np.array(f.olr.time.values, dtype='datetime64[D]')
    else:
        f = scipy.io.netcdf_file(str(filename), 'r')
        lat = f.variables['lat'].data.copy()
        lon = f.variables['lon'].data.copy()
        # scaling and offset as given in meta data of nc file
        olr = f.variables['olr'].data.copy() / 100. + 327.65
        hours_since1800 = f.variables['time'].data.copy()
        f.close()

        temptime = []
        for item in hours_since1800:
            delta = np.timedelta64(int(item / 24), 'D')
            day = np.datetime64('1800-01-01') + delta
            temptime.append(day)
        time = np.array(temptime, dtype=np.datetime64)
    result = OLRData(np.squeeze(olr), time, lat, lon)

    return result


def load_noaa_interpolated_olr_netcdf4(filename: Path, use_xarray: bool = False) -> OLRData:
    """
    Loads the standard OLR data product provided by NOAA in NetCDF4 format.

    ATTENTION: Whereas NetCDF4 seems to be the format of the recent NOAA OLR data files, the files originally used some
    years ago were saved as NetCDF3. So, if you are going to load the original files for reference purposes, you should
    use the loader function :py:func:`load_noaa_interpolated_olr` instead.

    The dataset can be obtained from
    https://www.psl.noaa.gov/thredds/catalog/Datasets/interp_OLR/catalog.html?dataset=Datasets/interp_OLR/olr.day.mean.nc

    A description is found at
    https://psl.noaa.gov/data/gridded/data.olrcdr.interp.html

    :param filename: Full filename of a local copy of OLR data file.
    :param use_xarray: Option to use xarray instead of NetCDF4 for faster loading of data and timestamps. If you consider
        activating this parameter, you should call instead :py:func:`load_noaa_interpolated_olr`, since it works for NetCDF
        3 and 4. Later on, when the xarray option is established, the present function particularly for NetCDF4 will be removed.
        Please also take into account the warnings in the docs of :py:func:`load_noaa_interpolated_olr`.
    :return: The OLR data.
    """
    if use_xarray:
        result = load_noaa_interpolated_olr(filename=filename, use_xarray=True)
    else:
        f = netcdf4.Dataset(filename, "r")
        lat = f.variables['lat'][:].data.copy()
        lon = f.variables['lon'][:].data.copy()
        olr = f.variables['olr'][:].data.copy()
        hours_since1800 = f.variables['time'][:].data.copy()
        f.close()

        temptime = []
        for item in hours_since1800:
            delta = np.timedelta64(int(item / 24), 'D')
            day = np.datetime64('1800-01-01') + delta
            temptime.append(day)
        time = np.array(temptime, dtype=np.datetime64)
        result = OLRData(np.squeeze(olr), time, lat, lon)

    return result


def restore_from_npzfile(filename: Path) -> OLRData:
    """
    Loads an :py:class:`OLRData` object from a numpy file, which has been saved with the function
    :py:func:`mjoindices.olr_handling.OLRData.save_to_npzfile`

    :param filename: The filename to the .npz file.

    :return: The OLR data.
    """
    with np.load(filename) as data:
        olr = data["olr"]
        time = data["time"]
        lat = data["lat"]
        long = data["long"]
    return OLRData(olr, time, lat, long)


def plot_olr_map_for_date(olr: OLRData, date: np.datetime64) -> Figure:
    """
    Plots a map pf the OLR data for a specific date.

    :param olr: The complete OLR data.
    :param date: The date for which the OLR data should be plotted
        (has to be exactly matched by a date of the OLR time grid).

    :return: The handle to the figure.
    """
    # TODO: Plot underlying map

    mapdata = olr.get_olr_for_date(date)

    if mapdata is not None:
        fig, axs = plt.subplots(1, 1, num="plot_olr_map_for_date", clear=True,
                                figsize=(10, 5), dpi=150, sharex=True, sharey=True)
        plt.subplots_adjust(wspace=0.35, hspace=0.35)

        ax = axs

        c = ax.contourf(olr.long, olr.lat, mapdata)
        fig.colorbar(c, ax=ax, label="OLR [W/m²]")
        ax.set_title("OLR")
        ax.set_ylabel("Latitude [°]")
        ax.set_xlabel("Longitude [°]")
    else:
        raise ValueError("No OLR data found for given date.")

    return fig




# -*- coding: utf-8 -*-

"""
Provides information for the installation routines.
"""

# Copyright (C) 2019 Christoph G. Hoffmann. All rights reserved.

# This file is part of mjoindices

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Contact: christoph.hoffmann@uni-greifswald.de

from setuptools import setup
import versioneer
import os

packages=['mjoindices', 'mjoindices.omi']

with open(os.path.join(os.path.dirname(__file__), 'README.md'), 'r') as f:
    long_description = f.read()

setup(name='mjoindices',
      packages=packages,
      package_dir={'': 'src'},
      #      package_data={'mjoindices': ['./src/mjoindices/omi/sign_reference/eof1/eof001.txt','./src/mjoindices/omi/sign_reference/eof2/eof001.txt'], },
      include_package_data=True,
      zip_safe=False,
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      install_requires=['numpy', 'pandas', 'scipy', 'matplotlib', 'netCDF4'],
      extras_require={
        'full_func': ['eofs', 'xarray'],
        'dev': ['build', 'pytest', 'pytest-pep8', 'tox', 'sphinx', 'm2r2', 'sphinx_rtd_theme'],
      },
      author='Christoph G. Hoffmann',
      author_email="christoph.hoffmann@uni-greifswald.de",
      url="https://github.com/cghoffmann/mjoindices",
      description="Calculation of indices that describe the Madden-Julian-Oscillation (only OMI by now)",
      long_description_content_type="text/markdown",
      long_description=long_description,
      license="GNU General Public License v3")
